# -*- coding: utf-8 -*-

import itertools

import numpy as np
import pandas as pd
from numpy.typing import ArrayLike
from scipy.spatial import ConvexHull

from simmate.apps.badelf.core.grid import Grid
from simmate.apps.badelf.core.partitioning import PartitioningToolkit
from simmate.toolkit import Structure


class VoxelAssignmentToolkit:
    """
    A set of tools for assigning charge to atoms in a unit cell.

    Args:
        charge_grid (Grid):
            A BadELF app Grid type object usually with CHGCAR type data.
        partitioning_grid (Grid):
            A BadELF app Grid type object usually with ELFCAR type data.
        algorithm (str):
            The algorithm to use for partitioning. Defaults to BadELF
        partitioning (dict):
            A partitioning dictionary generated from the BadELF app
            PartitioningToolkit. Will be generated from the grid if None.
        electride_structure (Structure):
            The structure with electride sites. Will be generated if not given.

    """

    def __init__(
        self,
        charge_grid: Grid,
        partitioning_grid: Grid,
        algorithm: str,
        electride_structure: Structure,
        partitioning: dict,
    ):
        self.partitioning_grid = partitioning_grid.copy()
        self.charge_grid = charge_grid.copy()
        self.algorithm = algorithm
        # partitioning will contain electride sites for voronelf
        self.partitioning = partitioning
        self.electride_structure = electride_structure

    @property
    def permutations(self):
        return self.charge_grid.permutations

    def get_matching_site(
        self,
        point_voxel_coord: ArrayLike | list,
        check_near_plane: bool = True,
    ):
        """
        Determines which atomic site a point belongs to.

        Args:
            point_voxel_coord (ArrayLike): The voxel coordinate of the point of
                interest.
            check_near_plane (bool): Whether or not to return None if the voxel
                might be intercepted by a plane.
        Returns:
            The site index that the point belongs to. Returns -1 if the point
            could be assigned to multiple sites.
        """
        partitioning = self.partitioning
        if check_near_plane:
            max_distance = self.charge_grid.max_voxel_dist
        else:
            # We don't want to check if a voxel is close to a plane. We make
            # the max distance negative so that when we check it's not possible
            # for the distance to be less than it.
            max_distance = -1
        # !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        # I've had a bug in the past where more than one site is found for a single
        # voxel. As such, I'm going to temporarily make this function search all
        # sites in case it finds more than one. This bug seemed to be due to incorrect
        # plane selection.
        # !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        sites = []
        # Iterate over each site in the lattice
        for site, neighbor_df in partitioning.items():
            matched = True
            # Iterate through each neighbor
            for neigh_index, row in neighbor_df.iterrows():
                # get the minimum point and normal vector which define the plane
                # seperating the sites.
                point = row["plane_points"]
                normal_vector = row["plane_vectors"]
                # If a voxel is on the same side of the plane as the site, then they
                # should have the same sign when their coordinates are plugged into
                # the plane equation (negative).
                # expected_sign = values["sign"]

                # use plane equation to find sign
                site_real_coord = self.charge_grid.get_cart_coords_from_vox(
                    point_voxel_coord
                )
                sign, distance = PartitioningToolkit.get_plane_sign(
                    point, normal_vector, site_real_coord
                )

                # if the sign matches, move to next neighbor.
                # if the sign ever doesn't match, then the site is wrong and we move
                # on to the next one after setting matched to false. We also check
                # to see if the voxel is possibly sliced by the plane. If it is we
                # want to seperate that charge later so we leave it here.
                if sign != "negative" or distance <= max_distance:
                    matched = False
                    break
            if matched == True:
                sites.append(site)
                # return site
        if len(sites) == 1:
            return sites[0]
        elif len(sites) == 0:
            return
        else:
            # Multiple sites found for one location
            return -1

    @property
    def electride_sites(self):
        """
        Function for getting the number of sites that are electrides

        Returns:
            A tuple of the indices corresponding to the electride sites in the
            structure.
        """
        structure = self.electride_structure
        if self.algorithm == "badelf":
            return structure.indices_from_symbol("He")
        elif self.algorithm == "voronelf":
            return []

    def get_voxels_site_base(
        self,
        point_voxel_coord: ArrayLike | list,
    ):
        """
        Finds the site a voxel belongs to. Mostly does the same as the
        get_matching_site function, but also checks other possible symmetric locations
        of each site.

        Args:
            point_voxel_coord (ArrayLike): The voxel coordinates of the point
                to find the associated site for.

            site (int): A site index. It is None if the voxel has not already
                been assigned to an electride.

        Returns:
            The site the voxel coordinate is associated with.
        """
        permutations = self.permutations
        # partitioning = self.partitioning
        x, y, z = point_voxel_coord
        # !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        # I've had a bug in the past where one voxel returns multiple sites after
        # being translated. I'm going to make this function temporarily go through
        # all sites in case the bug still exists. It will return -1 if the multiple
        # sites are found at the same transformation and it will return -2 if multiple
        # are found across different transformations. This bug seemed to be due to
        # insufficient partitioning plane selection
        # !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        sites = []
        # create dictionary for recording what fraction of a voxels volume should
        # be associated with a given site. We do this so that the format of the
        # output matches the other ones.
        site_vol_frac = {}
        for site in range(len(self.electride_structure)):
            site_vol_frac[site] = float(0)

        for t, u, v in permutations:
            new_voxel_coord = [x + t, y + u, z + v]
            site: int = self.get_matching_site(new_voxel_coord)
            # site returns none if no match, otherwise gives a number
            # The site can't return as an electride site as we don't include
            # electride sites in the partitioning results
            if site == -1:
                # If the site returns -1, this means multiple sites were found
                # at one transformation which should never happen. I want this
                # function to return -1 so I can count how often this happens
                # if at all.
                sites = []
                sites.append(-1)
                break
            elif site is not None:
                sites.append(site)
                # break
        if len(sites) > 1:
            # if the length of sites is greater than 1 that means it found more than
            # one site at different transformations. I want to return -2 so I can
            # keep track of how often this bug occurs.
            site_vol_frac[-2] = float(0)
            return site_vol_frac
        elif len(sites) == 1:
            # there is only one site found so we just return it.
            site_vol_frac[sites[0]] = float(1)
            return site_vol_frac
        else:
            site_vol_frac = None

            # there wasn't any site found so we don't return our site variable which
            # is None
        return site_vol_frac

    def _get_vertex_site(
        self,
        vertex_coords,
    ):
        """
        Finds the site a voxel vertex belongs to and returns the site and the
        transformation that it was found at. Mostly does the same as the
        get_voxels_site_base function, but also returns the transformation.

        Args:
            point_voxel_coord (ArrayLike): The voxel coordinates of the point
                to find the associated site for.

        Returns:
            The site a voxel vertex is assigned to as well as the transformation.
        """
        permutations = self.permutations
        x, y, z = vertex_coords

        # !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        # I've noticed a bug where one voxel returns multiple sites after
        # being translated. I'm going to make this function temporarily go through
        # all sites in case the bug still exists. It will return -1 if the multiple
        # sites are found at the same transformation and it will return -2 if multiple
        # are found across different transformations. This typically effects only
        # a small number of sites if the system is ionic. This bug seemed to be
        # caused by insufficient partitioning plane selection
        # !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

        sites = []
        translations = []
        for t, u, v in permutations:
            new_vert_coord = [x + t, y + u, z + v]
            site: int = self.get_matching_site(
                new_vert_coord,
                check_near_plane=False,
            )
            # site returns none if no match, otherwise gives a number
            # The site can't return as an electride site as we don't include
            # electride sites in the partitioning results
            if site == -1:
                # If the site returns -1, this means multiple sites were found
                # at one transformation which should never happen. I want this
                # function to return -1 so I can count how often this happens
                # if at all.
                sites = []
                sites.append(-1)
                break
            elif site is not None:
                sites.append(site)
                translations.append([t, u, v])
                # break
        if len(sites) > 1:
            # if the length of sites is greater than 1 that means it found more than
            # one site at different transformations. I want to return -2 so I can
            # keep track of how often this bug occurs.
            return -2, None
        elif len(sites) == 1:
            # there is only one site found so we just return it.
            if sites[0] == -1:
                return -1, None
            else:
                return sites[0], translations[0]
        else:
            # there wasn't any site found so we don't return our site variable which
            # is None
            return None, None
        # return site

    def get_voxel_vertices_sites(
        self,
        point_voxel_coord: ArrayLike | list,
    ):
        """
        Takes in the coordinates of a voxel and returns the coordinates of each
        of its vertices and the sites they belong to.

        Args:
            point_voxel_coord (ArrayLike): The voxel coordinates of the point
                to find the associated site for.
        """
        # create a dictionary to store the vertices coordinates and the site results
        vertices_coords = {}
        keys = ["A0", "A1", "A2", "A3", "B0", "B1", "B2", "B3"]
        # we need to transform the voxel coordinates to 8 different half coordinates
        # that represent its corners
        A0 = [-1 / 2, -1 / 2, -1 / 2]
        A1 = [-1 / 2, 1 / 2, -1 / 2]
        A2 = [1 / 2, 1 / 2, -1 / 2]
        A3 = [1 / 2, -1 / 2, -1 / 2]
        B0 = [-1 / 2, -1 / 2, 1 / 2]
        B1 = [-1 / 2, 1 / 2, 1 / 2]
        B2 = [1 / 2, 1 / 2, 1 / 2]
        B3 = [1 / 2, -1 / 2, 1 / 2]
        transforms = [A0, A1, A2, A3, B0, B1, B2, B3]
        for key, transform in zip(keys, transforms):
            vertices_coords[key] = [
                x + x1 for x, x1 in zip(point_voxel_coord, transform)
            ]

        # create a dataframe to store results in
        vertices_sites = pd.DataFrame(columns=["id", "transform", "site"])

        # now that we have the coordinates for each vertex, get the site and the
        # transform required to get the site
        for key, coord in vertices_coords.items():
            # try:
            site, transform = self._get_vertex_site(vertex_coords=coord)
            vertex_row = [key, transform, site]
            vertices_sites.loc[len(vertices_sites.index)] = vertex_row
            # If we can't find a site, return an empty row for this key
            # except:
            #     vertex_row = [key, None, None]
            #     vertices_sites.loc[len(vertices_sites.index)] = vertex_row
        # return both the vertices coords and the information about their sites
        return vertices_coords, vertices_sites

    @staticmethod
    def _get_vector_plane_intersection(
        point0: ArrayLike | list,
        point1: ArrayLike | list,
        plane_point: ArrayLike | list,
        plane_vector: ArrayLike | list,
        allow_point_intercept: bool = False,
    ):
        """
        Takes in two points and the point/vector defining a plane and returns
        the point where the line segment and plane intersect (if it exists)

        Args:
            point0 (ArrayLike): The first point of a line segment
            point1 (ArrayLike): The second point of a line segment
            plane_point (ArrayLike): A point on the plane
            plane_vector (ArrayLike): The vector normal to the plane
            allow_point_intercept (bool): Whether to count a point at the end
                of the line segment touching the plane as an intercept.

        Returns:
            The point where the line segment intersects the plane or None if
            there is no intersection.
        """
        # convert points to NumPy arrays
        point0 = np.array(point0)
        point1 = np.array(point1)
        plane_point = np.array(plane_point)
        plane_vector = np.array(plane_vector)

        # get direction of line segment.
        direction = point1 - point0
        # get dot product of direction vector and dot_product
        dot_product = np.dot(direction, plane_vector)
        # check if line is parallel to plane
        if np.abs(dot_product) < 1e-06:
            return None

        # get distance from the line segment point to the plane
        distance = np.dot(plane_point - point0, plane_vector) / dot_product

        # calculate intersection point
        intersection_point = point0 + direction * distance
        # round the intersection points

        # check if intersection point is between the start and end points of our
        # line segment. To do this, we would normally first check if the point
        # is on the line, but it must be alredy because we defined it as such. Next
        # we check the dot products
        # Get the vector for the intersecting point and point1 assuming point0 is
        # the origin
        AB = point1 - point0
        AC = intersection_point - point0
        # Get the dot products of the intersecting point and point1
        KAC = np.dot(AB, AC)
        KAB = np.dot(AB, AB)
        # There are five possible scenarios.
        if KAC < -1e-8 or KAC > KAB + 1e-8:
            # the point is outside the line
            return None
        elif KAC == 0 or KAC == KAB:
            # The point is at point0 or point1
            if allow_point_intercept:
                return intersection_point
            else:
                return None
        else:
            # The point is between point0 and point1
            return intersection_point

    def _get_voxel_plane_intersections(
        self,
        sites: list,
        vertices_coords: dict,
        intersections_df: pd.DataFrame,
    ):
        """
        Function for finding which planes intersect a voxel. Returns dataframe
        with all planes and intersections.

        Args:
            sites (list): A list of sites that the vertices of a voxel are assigned to.

            vertices_coords (dict): A dictionary relating the coordinates of a
                voxels vertices to their label.

            intersections_df (Dataframe): A dataframe of intersections that have
                already been found.

        Returns:
            A dataframe of where planes intersect the edges of a voxel.

        """
        partitioning = self.partitioning
        # Define the indices of the vertices making up the edges of the voxel.
        edges = [
            ["A0", "A1"],
            ["A1", "A2"],
            ["A2", "A3"],
            ["A3", "A0"],
            ["A0", "B0"],
            ["A1", "B1"],
            ["A2", "B2"],
            ["A3", "B3"],
            ["B0", "B1"],
            ["B1", "B2"],
            ["B2", "B3"],
            ["B3", "B0"],
        ]
        # We create a list of sites that we want to allow in the plane intersections
        # search. Each plane is stored twice: once for each atom in the pair. Because
        # of this we remove each site after we've looked at its planes to remove
        # redundancy
        sites_to_search = sites["site"].to_list()
        for site in sites["site"]:
            sites_to_search.remove(site)
            # iterate over each plane
            for neighbor_df_index, row in partitioning[site].iterrows():
                neighbor_index = row["neigh_index"]
                # if neighbor is in the list of sites, look at the plane
                if neighbor_index in sites_to_search:
                    plane_point = row["plane_points"]
                    plane_vector = row["plane_vectors"]
                    # iterate over each edge and find the points that intersect
                    intersections = []
                    for edge in edges:
                        point0 = vertices_coords[edge[0]]
                        point1 = vertices_coords[edge[1]]
                        intersection = self._get_vector_plane_intersection(
                            point0, point1, plane_point, plane_vector
                        )
                        if intersection is not None:
                            # we transform all intersections to be relative to A0 being
                            # the origin. This is so planes from various transformations
                            # can be treated at the same time
                            # MUST ROUND!
                            intersection = [
                                round((x - x1), 12)
                                for x, x1 in zip(intersection, vertices_coords["A0"])
                            ]
                            intersections.append(intersection)
                    if len(intersections) > 0:
                        # if we have any intersections we add the site index, its
                        # neighbors index, and the list of intersections to the df
                        plane_row = [site, row["neigh_index"], intersections]
                        intersections_df.loc[len(intersections_df)] = plane_row
        return intersections_df

    def get_intersected_voxel_volume_ratio(
        self,
        point_voxel_coord: ArrayLike | list,
    ):
        """
        Takes in the coordinates of a voxel and returns the ratio of the voxel
        that belongs to the sites around it.

        Args:
            point_voxel_coord (ArrayLike): The voxel coordinates of the point
                to find the associated site for.
        """
        grid = self.charge_grid.copy()
        voxel_volume = grid.voxel_volume
        vertices_coords, vertices_sites = self.get_voxel_vertices_sites(
            point_voxel_coord
        )

        # create dictionary for recording what fraction of a voxels volume should
        # be associated with a given site.
        site_vol_frac = {}
        for site in range(len(self.electride_structure)):
            site_vol_frac[site] = float(0)

        # shorten the lists to unique sites/planes
        sites = vertices_sites.drop_duplicates(subset="site", ignore_index=True)
        # !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        # I've made it so that if multiple sites are found for one vertex position,
        # it returns as -1. If multiple sites are found for one vertex but at multiple
        # transformed positions it returns -2.
        # My best guess for what is happening is that these sites are very close to
        # being exactly on a plane and are therefore returning as being part of
        # more than one site.
        # In a test with Na2S, returning these vertices as None and allowing the
        # program to continue gave more even results.
        if -1 in sites["site"].to_list() or -2 in sites["site"].to_list():
            # If vertices are found to have multiple sites I've made it so that it
            # records his information as a new site in the site_vol_frac dictionary.
            # This should be searchable in post so that I can keep count of these
            # without stopping the algorithm
            if -1 in sites["site"].to_list():
                site_vol_frac[-1] = float(0)
            elif -2 in sites["site"].to_list():
                site_vol_frac[-2] = float(0)
            sites = sites.replace(-1, None)
            sites = sites.replace(-2, None)
            # sites.replace()
            # breakpoint()
        # find the most common site for cases where only one site exists
        try:
            most_common_site = sites["site"].value_counts().idxmax()
        except:
            most_common_site = None
        #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

        sites = sites.dropna(subset="site").reset_index()

        # define the dataframe that we will store our plane information in
        intersections_df = pd.DataFrame(columns=["site", "neighbor", "intersections"])

        # Sometimes the voxel will be intersected only at certain transformations.
        # Often this is the permutation where the original voxel was able to be
        # assigned to an atomic site. We check this possibility first here. Sometimes,
        # the vertices themselves are shifted to other permutations. We check these
        # next.

        # Get the transform for the original site:
        # try:
        voxel_site, transform = self._get_vertex_site(
            point_voxel_coord,
        )
        # except:
        #     voxel_site, transform = None, None

        # If no intersections are found, check the other possible transforms
        if len(intersections_df) == 0:
            # Get the list of possible transforms for the dataframe of sites and
            # transforms we found earlier
            transforms = vertices_sites.drop_duplicates(subset="transform").dropna()
            transforms = transforms["transform"].to_list()

            # Now iterate over the remaining transforms to find locations where the
            # edge is intersected.
            for transform in transforms:
                transformed_coords_real = {}
                for key, coord in vertices_coords.items():
                    # get the transformed coordinate, convert it to a real coordinate,
                    # and add to our new dictionary of vertices
                    new_coord = [x + t for x, t in zip(coord, transform)]
                    real_coord = self.charge_grid.get_cart_coords_from_vox(new_coord)
                    transformed_coords_real[key] = real_coord
                # get any intersections between planes and voxel edges and add to
                # our intersection dataframe
                intersections_df = self._get_voxel_plane_intersections(
                    sites, transformed_coords_real, intersections_df
                )
        # It's possible for the plane to intersect a vertex exactly causing there
        # to be multiple instances of the same intersection. Remove any duplicates
        # this may cause.
        intersections_df = intersections_df.drop_duplicates(subset="intersections")

        # We shift all intersections to the origin in case there are multiple
        # planes intersecting at different transforms. We also need to shift all
        # of the vertices so that they are relative to the origin.
        vertices_coords_real_origin = {}
        # convert coordinate to real for use later
        for key, coord in vertices_coords.items():
            # transform so that A0 is at origin
            transformed_coord = [
                x - x1 + 1 for x, x1 in zip(coord, vertices_coords["A0"])
            ]
            # add the real coord to our dictionary of vertices
            vertices_coords_real_origin[key] = grid.get_cart_coords_from_vox(
                transformed_coord
            )

        # Now that we have a list of plane intersections, we can find the what portion
        # of the voxels should be applied to each site.
        # If there is only one site in the list, return the site fraction dictionary
        # with 1 for this site.
        if len(sites) == 1:
            if voxel_site is not None:
                #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                # Sometimes the voxel center is returned as belonging to more than
                # 1 site. I beleive this is because without the condition that the
                # voxel needs to be some distance away from the plane (Which exists
                # when we assign planes earlier in the code.) it is actually possible
                # for a plane to exactly go through the site resulting in it
                # being allowed to belong to two sites. In this situation I don't want
                # to give the voxel to a -2 site, I want it to go to the most common
                # site.
                if voxel_site == -1 or voxel_site == -2:
                    try:
                        site_vol_frac[most_common_site] = 1.0
                    except:
                        site_vol_frac = None
                #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                elif len(intersections_df) == 0 or len(intersections_df) == 1:
                    site_vol_frac[voxel_site] = 1.0
                else:
                    site_vol_frac = None
            else:
                site_vol_frac = None
        # If there are two sites, at least one plane is intersecting the voxel
        elif len(sites) == 2:
            # if there are not intersections found, then something is not working
            if len(intersections_df) == 0:
                #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                # I'm having this problem return as -3 so that I can keep track of it
                site_vol_frac[-3] = float(0)
            #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
            # Check that there is only one plane that is close to all of the vertices.
            # If so, our voxel is being split by a single plane.
            elif len(intersections_df) == 1:
                # get a list of the intersecting points
                intersect_points = list(intersections_df["intersections"])[0]

                # If there are no intersect points, there is an error somewhere
                # above and we want to stop the process.
                if len(intersect_points) == 0:
                    breakpoint()
                else:
                    # define a list to populate with the points that make up one section
                    # of the voxel
                    site_points = []
                    # define the two sites that the voxel will be split into
                    site1 = sites["site"].iloc[0]
                    site2 = sites["site"].iloc[1]
                    # iterate over the vertices and if they belong to the first site
                    # append them to our site points list
                    for row in vertices_sites.iterrows():
                        if row[1]["site"] == site1 or row[1]["site"] is None:
                            site_points.append(
                                np.array(vertices_coords_real_origin[row[1]["id"]])
                            )
                        # otherwise we don't do anthing
                    # combine the list of vertex coordinates with the list of intersections

                    hull_points = []
                    # Add all vertex points to the list
                    for site_point in site_points:
                        hull_points.append(np.array(site_point))
                    # If intersect points are found very close to a vertex point, we
                    # don't want to accidentally double count. Here we check each
                    # point before adding it to the list of hull points.
                    for intersect_point in intersect_points:
                        intersect_array = np.array(intersect_point)
                        repeat = False
                        for site_point in site_points:
                            site_array = np.array(site_point)
                            if np.allclose(
                                intersect_array, site_array, rtol=0, atol=1e-6
                            ):
                                repeat = True
                                break
                            else:
                                pass
                        if repeat == False:
                            hull_points.append(intersect_array)

                    # define a 3D convex hull for the first segment. Get its volume, then
                    # use the ratio of its volume to the total voxel volume to find the
                    # ratio of charge that should be applied to each voxel
                    try:
                        hull = ConvexHull(hull_points)
                        seg1_vol = hull.volume
                        seg1_ratio = round((seg1_vol / voxel_volume), 16)
                        site_vol_frac[site1] = seg1_ratio
                        site_vol_frac[site2] = round((1 - seg1_ratio), 16)
                    except:
                        print(f"Error with hull: {point_voxel_coord}")
                        site_vol_frac = None

            elif len(intersections_df) > 1:
                # if there is more than one plane, the voxel is being split in
                # multiple ways. For now we just return no site_vol_frac because
                # we won't be handling this rigorously
                site_vol_frac = None
        # if there are more than two sites or more than one planes, the voxel is
        # being split by more than one plane. For now I'm just passing these on
        else:
            site_vol_frac = None
        return site_vol_frac

    def get_voxels_multi_plane(
        self,
        point_voxel_coords: ArrayLike | list,
        site_vol_frac: dict,
        voxel_assignments: pd.DataFrame,
    ):
        """
        This function finds what sites a voxel divided by more than one plane should
        be applied to. It looks at nearby voxels and finds what sites they are applied
        to and finds the ratio between these sites.
        """
        if site_vol_frac is None:
            x, y, z = point_voxel_coords
            grid_shape = self.charge_grid.grid_shape
            electride_sites = self.electride_sites
            # create dictionary for counting number of sites and for the fraction
            # of sites
            site_count = {}
            site_vol_frac = {}
            for site in range(len(self.electride_structure)):
                site_count[site] = 0
                site_vol_frac[site] = 0
            # look at all neighbors around the voxel of interest
            for t, u, v in itertools.product([-1, 0, 1], [-1, 0, 1], [-1, 0, 1]):
                new_idx = [x - 1 + t, y - 1 + u, z - 1 + v]

                # wrap around for voxels on edge of cell
                new_idx = [a % b for a, b in zip(new_idx, grid_shape)]

                # get site ditionary from the neighboring voxel using its row
                # index. This is much faster than searching by values. To get the index
                # we can utilize the fact that an increase in z will increase the index
                # by 1, an increase in y will increase the index by (range of z),
                # and an increase in x will increase the index by (range of z)*(range of y)
                zrange = grid_shape[2]
                yrange = grid_shape[1]
                index = int(
                    (new_idx[0]) * zrange * yrange + (new_idx[1]) * zrange + new_idx[2]
                )
                site_dict = voxel_assignments["site"].iloc[index]
                # If the site dict exists, we want to look at each of its key/item
                # pairs. -1,-2,-3 indicate some error in voxel assignment. We skip
                # these for now. Also, if the algorithm is "badelf" we don't want to
                # assign anything to electride sites so we also sip thee. Otherwise
                # we add the fraction of each voxel to the voxel count
                #!!! avoiding electrides is going to cause problems for voronelf
                if site_dict is not None:
                    for site_index, frac in site_dict.items():
                        if site_index in [-3, -2, -1]:
                            continue
                        elif (
                            site_index in electride_sites and self.algorithm == "badelf"
                        ):
                            continue
                        else:
                            site_count[site_index] += frac

            if sum(site_count.values()) != 0:
                # get the fraction of each site. This will allow us to split the
                # charge of the voxel more evenly
                for site_index in site_count:
                    site_vol_frac[site_index] = site_count[site_index] / sum(
                        site_count.values()
                    )

            else:
                #  of the voxels nearby returned either None, as an error, or as
                # an electride. To assign this voxel we just give it to the closest
                # atom
                site_vol_frac = self.get_voxels_site_nearest(point_voxel_coords)
        return site_vol_frac

    def get_voxels_site_nearest(
        self,
        point_voxel_coords: ArrayLike | list,
    ):
        if self.algorithm == "badelf":
            structure = self.charge_grid.structure
        elif self.algorithm == "voronelf":
            structure = self.electride_structure
        structure_temp = structure.copy()
        structure_temp.append("He", point_voxel_coords, coords_are_cartesian=True)
        nearest_site = structure_temp.get_neighbors(structure_temp[-1], 5)[0].index
        # create dictionary for recording what fraction of a voxels volume should
        # be associated with a given site.
        site_vol_frac = {}
        for site in range(len(self.electride_structure)):
            site_vol_frac[site] = float(0)
        site_vol_frac[nearest_site] = float(1)

        return site_vol_frac

    def get_voxels_site(
        self,
        point_voxel_coords: ArrayLike | list,
        site_vol_frac,
    ):
        """
        Combines methods of voxel assignment to guarantee that the voxel
        site is returned.
        """
        if site_vol_frac is None:
            site_vol_frac = self.get_voxels_site_base(point_voxel_coords)
        if site_vol_frac is None:
            site_vol_frac = self.get_intersected_voxel_volume_ratio(point_voxel_coords)

        return site_vol_frac

    def _get_voxels_site_dask(
        self,
        voxel_dataframe: pd.DataFrame,
    ):
        """
        Applies the get_intersected_voxel_volume_ratio function across a Pandas
        dataframe. This is the function that we give to dasks map_partitions
        function to parallelize voxel assignment.

        Args:
            voxel_dataframe (pd.DataFrame): A dataframe consisting of columns
            [x,y,z] representing the voxels of a system.

        Returns:
            A series of dictionaries containing the ratio of each voxel assigned
            to each site.
        """
        return voxel_dataframe.apply(
            lambda x: self.get_voxels_site(
                [x["x"], x["y"], x["z"]],
                x["site"],
            ),
            axis=1,
        )



# -*- coding: utf-8 -*-

from functools import cache  # cached_property doesnt work with classmethod

try:
    from prefect.client import get_client
    from prefect.context import FlowRunContext
    from prefect.flows import Flow
    from prefect.orion.schemas.filters import FlowFilter, FlowRunFilter
    from prefect.packaging import OrionPackager
    from prefect.packaging.serializers import PickleSerializer
    from prefect.states import State
    from prefect.tasks import task  # present only for convience imports elsewhere

    from simmate.utilities.async_wrapper import async_to_sync
except:
    raise Exception(
        "Please install prefect with `conda install -c conda-forge prefect anyio`. "
        "All prefect features will be broken until this is fixed"
    )


class PrefectWorkflow:
    @classmethod
    def run(cls, **kwargs) -> State:
        """
        runs the workflow locally in a prefect context
        """

        subflow = cls._to_prefect_flow()
        state = subflow(return_state=True, **kwargs)

        # We don't want to block and wait because this might disable parallel
        # features of subflows. We therefore return the state and let the
        # user decide if/when to block.
        # result = state.result()

        return state

    @classmethod
    def run_cloud(cls, return_future: bool = True, **kwargs) -> str:
        """
        This schedules the workflow to run remotely on Prefect Cloud.

        #### Parameters

        - `labels`:
            a list of labels to schedule the workflow with

        - `wait_for_run`:
            whether to wait for the workflow to finish. If False, the workflow
            will simply be submitted and then exit. The default is True.

        - `**kwargs`:
            all options that are normally passed to the workflow.run() method

        #### Returns

        - The flow run id that was used in prefect cloud.


        #### Usage

        Make sure you have Prefect properly configured and have registered your
        workflow with the backend.

        Note that this method can be viewed as a fork of:
            - from prefect.tasks.prefect.flow_run import create_flow_run
        It can also be viewed as a more convenient way to call to client.create_flow_run.
        I don't accept any other client.create_flow_run() inputs besides 'labels'.
        This may change in the future if I need to set flow run names or schedules.
        """

        # Prefect does not properly deserialize objects that have
        # as as_dict or to_dict method, so we use a custom method to do that here
        parameters_serialized = cls._serialize_parameters(**kwargs)
        # BUG: What if we are submitting using a filename? We don't want to
        # submit to a cluster and have the job fail because it doesn't have
        # access to the file. One solution could be to deserialize right before
        # serializing in the next line in order to ensure parameters that
        # accept file names are submitted with all necessary data.

        # Now we submit the workflow.
        run_id = cls._submit_to_prefect(parameters=parameters_serialized)

        cls._register_calculation(run_id=run_id, **kwargs)
        # BUG: Will there be a race condition here? What if the workflow finishes
        # and tries writing to the databse before this is done?
        # BUG: if parameters are improperly set, this line will fail, while the
        # job submission (above) will suceed. Should I cancel the flow run if
        # this occurs?

        if return_future:
            raise Exception("Prefect cannot return futures from submisson.")

        # return the flow run_id for the user
        return run_id

    @classmethod
    @property
    def tags(cls) -> list[str]:
        """
        Lists of tags to submit a the workflow with when using run_cloud.
        """
        return [
            "simmate",
            cls.name_type,
            cls.name_app,
        ]

    @staticmethod
    def _get_run_id(raise_if_no_context: bool = True):
        """
        grabs the prefect flow run id from context (if there is one)
        """
        # Grab the flow run id for reference.
        run_context = FlowRunContext.get()
        if run_context:
            run_id = str(run_context.flow_run.id)
            return run_id
        elif not run_context and not raise_if_no_context:
            return None  # no error is raised
        else:
            raise Exception("Cannot detetect a Prefect FlowRunContext")

    @classmethod
    @cache
    def _to_prefect_flow(cls) -> Flow:
        """
        Converts this workflow into a Prefect flow
        """

        # Instead of the @flow decorator, we build the flow instance directly
        flow = Flow(
            fn=cls._run_full,
            name=cls.name_full,
            version=cls.version,
            # Skip type checking because I don't have robust typing yet
            # e.g. Structure type inputs also accept inputs like a filename
            validate_parameters=False,
        )

        # as an extra, we set this attribute to the prefect flow instance, which
        # allows us to access the source Simmate Workflow easily with Prefect's
        # context managers.
        flow.simmate_workflow = cls

        return flow

    @classmethod
    @property
    @cache
    @async_to_sync
    async def depolyment_id_prefect(cls) -> str:
        """
        Grabs the deployment id from the prefect database if it exists, and
        if not, creates the depolyment and then returns the id.

        This is a synchronous and cached version of `_get_depolyment_id_prefect` and
        this is the preferred method to use for beginners.
        """
        return await cls._get_depolyment_id_prefect()

    @classmethod
    async def _get_depolyment_id_prefect(cls) -> str:
        """
        Grabs the deployment id from the prefect database if it exists, and
        if not, creates the depolyment and then returns the id.

        This is an asynchronous method and should only be used when within
        other async methods. Beginners should instead use the `depolyment_id_prefect`
        property.
        """

        async with get_client() as client:
            response = await client.read_deployments(
                flow_filter=FlowFilter(
                    name={"any_": [cls.name_full]},
                ),
            )

        # If this is the first time accessing the deployment id, we will need
        # to create the deployment
        if not response:
            deployment_id = await cls._create_deployment_prefect()

        # there should only be one deployment associated with this workflow
        # if it's been deployed already.
        elif len(response) == 1:
            deployment_id = str(response[0].id)

        else:
            raise Exception("There are duplicate deployments for this workflow!")

        return deployment_id

    @classmethod
    async def _create_deployment_prefect(cls) -> str:
        """
        Registers this workflow to the prefect database as a deployment.

        This method should not be called directly. It will be called by
        other methods when appropriate
        """

        # raise error until python-deployments are supported again
        raise Exception(
            "Prefect 2.0 has removed the ability to create deployments in "
            "python, so this feature is currently disabled."
        )
        # When this is removed, be sure to re-add the test_workflow_cloud unittest
        # UPDATE: looks like they re-added it in 2.1.0. You'd use the
        # build_from_flow method:
        # https://docs.prefect.io/api-ref/prefect/deployments/

        from prefect.deployments import Deployment

        # NOTE: we do not use the client.create_deployment method because it
        # is called within the Deployment.create() method for us.
        deployment = Deployment(
            name=cls.name_full,
            flow=cls._to_prefect_flow(),
            packager=OrionPackager(serializer=PickleSerializer()),
            # OPTIMIZE: it would be better if I could figure out the ImportSerializer
            # here. Only issue is that prefect would need to know to import AND
            # call a method.
            tags=cls.tags,
        )

        deployment_id = await deployment.create()

        return str(deployment_id)  # convert from UUID to str first

    @classmethod
    @async_to_sync
    async def _submit_to_prefect(cls, **kwargs) -> str:
        """
        Submits a flow run to prefect cloud.

        This method should not be used directly. Instead use `_run_prefect_cloud`.
        """

        # The reason we have this code as a separate method is because we want
        # to isolate Prefect's async calls from Django's sync-restricted calls
        # (i.e. django raises errors if called within an async context).
        # Therefore, methods like `_run_prefect_cloud` can't have both this async code
        # AND methods like _register_calculation that make sync database calls.

        async with get_client() as client:
            response = await client.create_flow_run_from_deployment(
                deployment_id=await cls._get_depolyment_id_prefect(),
                **kwargs,
            )

        flow_run_id = str(response.id)
        return flow_run_id

    @classmethod
    @property
    @async_to_sync
    async def nflows_submitted(cls) -> int:
        """
        Queries Prefect to see how many workflows are in a scheduled, running,
        or pending state.
        """

        async with get_client() as client:
            response = await client.read_flow_runs(
                flow_filter=FlowFilter(
                    name={"any_": [cls.name_full]},
                ),
                flow_run_filter=FlowRunFilter(
                    state={"type": {"any_": ["SCHEDULED", "PENDING", "RUNNING"]}}
                ),
            )

        return len(response)



# -*- coding: utf-8 -*-

from simmate.apps.warren_lab.workflows.relaxation.pbesol import (
    Relaxation__Vasp__WarrenLabPbesol,
)
from simmate.apps.warren_lab.workflows.static_energy.pbe import PBE_STATIC_SETTINGS


class StaticEnergy__Vasp__WarrenLabPbesol(Relaxation__Vasp__WarrenLabPbesol):
    """
    Performs a static energy calculation based on the settings for Warren Lab
    PBEsol functional relaxation. This functional is generally considered to
    be more accurate for solids.
    (Phys. Rev. Lett. 102, 039902 (2009))
    """

    _incar_updates = PBE_STATIC_SETTINGS



# -*- coding: utf-8 -*-

from .relaxation import (
    Relaxation__Vasp__Quality00,
    Relaxation__Vasp__Quality01,
    Relaxation__Vasp__Quality02,
    Relaxation__Vasp__Quality03,
    Relaxation__Vasp__Quality04,
    Relaxation__Vasp__Staged,
)
from .static_energy import StaticEnergy__Vasp__Quality04



# -*- coding: utf-8 -*-

from pymatgen.symmetry.analyzer import SpacegroupAnalyzer

from simmate.toolkit import Structure
from simmate.toolkit.transformations.base import Transformation


class ExtremeSymmetry(Transformation):
    """
    Applies a series of increasingly excessive symmetry checks -- starting
    at 0.1 Angstrom + 0.5 Deg and ending at ~3 Angstrom + ~40 Deg. The final
    tolerance checks are massive tolerances that are intended to reduce the
    structure to a new and distinct one.

    This will return the highly symmetrized structure, which is often useful
    for evolutionary searches where many disordered structures exist. We expect
    it to be most effective for structures with high site counts.
    """

    ninput = 1
    io_scale = "one_to_one"
    allow_parallel = True

    @staticmethod
    def apply_transformation(structure: Structure):
        # To establish the original structure symmetry, we begin with strict
        # tolerances of 0.01Angstrom and 0.5 Deg cutoffs
        sga = SpacegroupAnalyzer(
            structure,
            symprec=0.01,
            angle_tolerance=0.5,
        )
        original_spacegroup = sga.get_space_group_number()

        # As we try higher tolerances, we want to keep track of when we find
        # a new & higher spacegroup.
        results = {original_spacegroup: [0.01, 0.5]}
        symprec = 0.1
        angle_tol = 0.5

        # This cycle attempts symmetry reduction at different tolerances.
        # The tolerances increase as we go, and no matter what, 24 different
        # attempts will be made.
        for attempt in range(25):
            # start-up the new analysis
            sga = SpacegroupAnalyzer(
                structure,
                symprec=symprec,
                angle_tolerance=angle_tol,
            )

            # For extremely high tolerances, spglib will return None and cause
            # pymatgen to fail with an error. We want to simply move on if
            # an error happens.
            try:
                new_spacegroup = sga.get_space_group_number()
            except TypeError:
                continue

            # We only store the results if we generated a brand-new spacegroup
            if new_spacegroup not in results.keys():
                results.update({new_spacegroup: sga.get_primitive_standard_structure()})

            # For the next cycle, I simple scale the angle and distance tolerances
            # by a fixed percent. This results in the following overall checks:
            #   Ang: 0.11499999999999999, DEG: 0.6
            #   Ang: 0.13224999999999998, DEG: 0.72
            #   Ang: 0.15208749999999996, DEG: 0.864
            #   Ang: 0.17490062499999995, DEG: 1.0368
            #   Ang: 0.20113571874999991, DEG: 1.24416
            #   Ang: 0.23130607656249988, DEG: 1.4929919999999999
            #   Ang: 0.26600198804687486, DEG: 1.7915903999999998
            #   Ang: 0.3059022862539061, DEG: 2.1499084799999997
            #   Ang: 0.35178762919199197, DEG: 2.5798901759999997
            #   Ang: 0.40455577357079076, DEG: 3.0958682111999996
            #   Ang: 0.46523913960640934, DEG: 3.7150418534399994
            #   Ang: 0.5350250105473707, DEG: 4.458050224127999
            #   Ang: 0.6152787621294763, DEG: 5.349660268953598
            #   Ang: 0.7075705764488976, DEG: 6.419592322744317
            #   Ang: 0.8137061629162322, DEG: 7.7035107872931805
            #   Ang: 0.935762087353667, DEG: 9.244212944751816
            #   Ang: 1.0761264004567168, DEG: 11.093055533702179
            #   Ang: 1.2375453605252242, DEG: 13.311666640442615
            #   Ang: 1.4231771646040077, DEG: 15.973999968531137
            #   Ang: 1.6366537392946088, DEG: 19.168799962237365
            #   Ang: 1.8821518001888, DEG: 23.002559954684838
            #   Ang: 2.16447457021712, DEG: 27.603071945621803
            #   Ang: 2.489145755749688, DEG: 33.12368633474616
            #   Ang: 2.8625176191121406, DEG: 39.74842360169539
            symprec *= 1.15
            angle_tol *= 1.2
            # print(f"Ang: {symprec}, DEG: {angle_tol}") --> generates comment above

        # Grab the final & highest symmetry result.
        max_spacegroup = max(list(results.keys()))
        final_structure = results[max_spacegroup]
        # TODO: in the future I could make this a "one_to_many" transformation
        # and return all of the higher-symmetry structures

        # make sure we actually improved the symmetry and didn't just grab
        # the original structure
        if max_spacegroup == original_spacegroup:
            # Failed to generate a higher-symmetry structure
            return

        return final_structure



# -*- coding: utf-8 -*-

import logging

from django.apps import apps
from django.core.management import call_command
from django.db.utils import DatabaseError

from simmate.configuration import settings

# Lists off which apps to update/create. By default, I do all apps that are installed
# so this list is grabbed directly from django. I also grab the CUSTOM_APPS to
# check for user-installed applications.
APPS_TO_MIGRATE = list(apps.app_configs.keys())


def check_db_conn(original_function: callable):
    """
    A decorator that catches errors such as "close connection" failures and
    retries with a new connection.

    ## Example use:
    ``` python
    @check_db_conn
    def example():
        return 12345 # some fxn that makes database calls
    ```
    """
    # BUGFIX: for processes (e.g. workflows) that take >1hr, the database
    # connection to postgres can be dropped/terminated. So we need to catch
    # this and make a new connection.
    #   https://github.com/jacksund/simmate/issues/364

    # New feature in Django worth exploring if this becomes an issue again.
    # However, this is only for web views... not local dev/runs:
    #   https://docs.djangoproject.com/en/4.1/ref/settings/#conn-health-checks

    def wrapper(*args, **kwargs):
        # On our first try, just use default method and existing connection
        try:
            return original_function(*args, **kwargs)

        # This 2nd attempt is an exact retry where we grab a new db connection
        # Fix is from:
        #   https://stackoverflow.com/questions/48329685
        except DatabaseError as error:
            logging.critical(error)
            logging.info("retrying with new db connection")

            # Note, this import needs to be done locally! Having it imported
            # above causes pickling errors.
            #   see https://github.com/jacksund/simmate/issues/410
            from django.db import connection as db_connection

            db_connection.connect()

            # retry the function call
            return original_function(*args, **kwargs)

    return wrapper


def update_database(
    apps_to_migrate: list[str] = APPS_TO_MIGRATE,
    show_logs: bool = True,
):
    # check Django if there are any updates to be made
    if show_logs:
        logging.info("Checking for and applying updates...")

    # execute the following commands to update the database
    call_command("makemigrations", *apps_to_migrate)
    call_command("migrate")

    # Let the user know everything succeeded
    if show_logs:
        logging.info("Success! Your database tables are now up to date. :sparkles:")


def reset_database(
    apps_to_migrate: list[str] = APPS_TO_MIGRATE,
    use_prebuilt: bool = False,
):
    # TODO: call_command("flush") could be used in the future to simply
    # delete all data -- without rerunning migrations

    # We can now proceed with reseting the database
    logging.info("Removing database and rebuilding...")

    # BUG: this is only for SQLite3 and Postgres
    # If I wish to add FULL functionality of all DBs, I could consider
    # wrapping the django-extensions function for this instead:
    #   https://django-extensions.readthedocs.io/en/latest/reset_db.html
    # An example command to call this (when django-extensions is installed) is...
    #   django-admin reset_db --settings=simmate.configuration.django.settings
    # Note: this does not remove migration files or reapply migrating after

    if settings.database_backend == "sqlite3":
        # grab the location of the database file. I assume the default
        # database for now.
        db_filename = settings.database.name

        # delete the sqlite3 database file if it exists
        if db_filename.exists():
            db_filename.unlink()

    elif settings.database_backend == "postgresql":
        # We do this with an independent postgress connection, rather than through
        # django so that we can close everything down easily.
        import psycopg2

        # Setup Postgres connection
        # Postgres requires a 'maintenance database' that we connect to while
        # we add/drop the table. For most cases, such as building a new database
        # through PgAdmin or a Docker image, the maintenance database will be
        # named "postgres". For Digitial Ocean, it will be named "defaultdb".
        # As a last resort, we can also check if there is a database matching
        # the name of the user. We iterate through these common cases until
        # we find one that works, and then warn the user if things fail.
        connection = None
        for maintenance_db_name in [
            "postgres",
            "defaultdb",
            settings.database.user,
        ]:
            try:
                connection = psycopg2.connect(
                    host=settings.database.host,
                    database=maintenance_db_name,
                    user=settings.database.user,
                    password=settings.database.password,
                    port=settings.database.port,
                )
            except psycopg2.OperationalError as error:
                if f'"{maintenance_db_name}" does not exist' in str(error):
                    continue  # just jump to trying the next db name
                # otherwise we might a password auth issue or something else
                raise error

            # exit loop as soon as we have a working connection
            if connection:
                break

        # ensure the loop above found a working connection
        if not connection:
            raise Exception(
                "Postgres requires a 'maintenance database' that we connect to "
                "while we add/drop/reset the database that you'd like to use. "
                "That database can stay empty, but it's important to be present. "
                "Simmate was unable to detect your maintenance database, which "
                "is why you're seeing this error. To fix this, make sure you have "
                "a database named either 'postgres', 'defaultdb', or one that has "
                "an identical name to your username. Create this database on your "
                "postgres server with a SQL command such as 'CREATE DATABASE "
                "defaultdb' and then retry your simmate command."
            )

        # In order to delete a full database, we need to isolate this call
        connection.set_isolation_level(0)

        # Open connection cursor to perform database operations
        cursor = connection.cursor()

        # Build out database extensions and tables
        db_name = settings.database.name
        cursor.execute(f'DROP DATABASE IF EXISTS "{db_name}";')
        # BUG: if others are connected I could add 'WITH (FORCE)' above.
        # For now, I don't use this but should consider adding it for convenience.
        # I think this is buggy with older versions of postgres (like RDkit),
        # so I hold off on this for now.
        cursor.execute(f'CREATE DATABASE "{db_name}";')

        # Make the changes to the database persistent
        connection.commit()

        # Close communication with the database
        cursor.close()
        connection.close()

    elif settings.database_backend not in ["postgresql", "sqlite3"]:
        logging.warning(
            "reseting your database is only supported for SQLite and Postgres."
            " Make sure you only use this function when initially building your "
            "database and not after."
        )

    # instead of building the database from scratch, we instead download a
    # prebuilt database file.
    if settings.database_backend == "sqlite3" and use_prebuilt:
        from simmate.database.third_parties import load_default_sqlite3_build

        logging.info("Setting up prebuilt database...")
        load_default_sqlite3_build()

        # now update the database based on the registered apps
        update_database(apps_to_migrate, show_logs=False)

    # Otherwise we make an empty database.
    # Because this is our first time building the database, we also want to
    # load the Spacegroup metadata for us to query Structures by.
    else:
        from simmate.database.base_data_types import Spacegroup

        logging.info("Building empty database...")
        update_database(apps_to_migrate, show_logs=False)

        logging.info("Loading default data...")
        Spacegroup._load_database_from_toolkit()

    # Let the user know everything succeeded
    logging.info("Success! Your database has been reset. :sparkles:")


def dump_database_to_json(
    filename: str = "database_dump.json",
    exclude: list[str] = [],
):
    # BUG: https://stackoverflow.com/questions/67616945/
    # os.environ["PYTHONIOENCODING"] = "utf8"  # DOESNT WORK...
    # python -Xutf8 manage.py dumpdata > data.json  # WORKS!

    # Begin writing the database to the json file.
    logging.info("Writing all data to JSON...")

    # execute the following commands to write the database to a json file
    call_command("dumpdata", output=filename, exclude=exclude)

    # Let the user know everything succeeded
    logging.info(
        f"Success! You should now see the file {filename} with all of your data."
    )


def load_database_from_json(filename: str = "database_dump.json"):
    # Begin writing the database to the json file.
    logging.info("Loading all data from JSON...")

    # OPTIMIZE: this function is very slow. Consider speed-up options such as
    # making this function a transaction or manually writing a bulk_create. It
    # actually looks like django ORM takes up most of the time tough, and the actual
    # database queries are not the bottleneck...

    # execute the following commands to build the database
    # BUG: contenttypes gives issues because a migrated database already has these
    # set. Simply ignore this table and everything works. The contenttypes is
    # simply a table that lists all of our different models.
    call_command("loaddata", filename, exclude=["contenttypes"])

    # Let the user know everything succeeded
    logging.info(
        f"Success! You now have all the data from {filename} available in your database."
    )


# BUG: This function isn't working as intended
# def graph_database(filename="database_graph.png"):
#     # using django-extensions, we want to make an image of all the available
#     # tables in our database as well as their relationships.
#     # This is the equivalent of running the following command:
#     #   django-admin graph_models -a -o image_of_models.png --settings=...
#     call_command("graph_models", output=filename, all_applications=True, layout="fdp")



# -*- coding: utf-8 -*-

from .pwscf_xml import PwscfXml

# -----------------------------------------------------------------------------

# DEV NOTES:

# -- pwscf.out --
# This file is not read or parsed anywhere
# This is because the `pwscf.out` file contains minimal information & not all
# of it is in a machine readable format. All information (and more) is in
# the `pwscf.xml` file instead, which we use instead.
# If we ever want to parse the out file instead, we can fork pymatgen's parser:
#   https://github.com/materialsproject/pymatgen/blob/master/pymatgen/io/pwscf.py


# -- pwscf.save/*.dat files -- (e.g. charge-density.dat)
# These files are in binary format and can only be read on the OS they were
# created/written on. This causes trouble for file sharing... Ideally we can
# convert these to XML (using FoX) before trying to read. Otherwise, binary
# files can be loaded using h5py.
#   https://github.com/QEF/qeschema/blob/master/qeschema/hdf5/__init__.py#L15

# -----------------------------------------------------------------------------



# -*- coding: utf-8 -*-

from django.shortcuts import render

from simmate.configuration import settings
from simmate.database.base_data_types import DatabaseTable, Structure
from simmate.website.core_components.base_api_view import SimmateAPIViewSet

# closed-source data types. We can skip these if they aren't present
try:
    from simmate_corteva.rdkit.models import Molecule
except ModuleNotFoundError:
    Molecule = None  # dummy class just for comparison below


ALL_PROVIDERS = {
    DatabaseTable.get_table(table_name).table_name: DatabaseTable.get_table(table_name)
    for table_name in settings.website.data
}


def providers_all(request):
    crystal_dbs = []
    molecular_dbs = []
    other_dbs = []
    for table in ALL_PROVIDERS.values():
        if issubclass(table, Structure):
            crystal_dbs.append(table)
        elif Molecule and issubclass(table, Molecule):
            molecular_dbs.append(table)
        else:
            other_dbs.append(table)

    context = {
        "all_datasets": {
            "Molecular": molecular_dbs,
            "Crystalline": crystal_dbs,
            "Other": other_dbs,
        }
    }
    template = "data_explorer/providers_all.html"
    return render(request, template, context)


class ProviderAPIViewSet(SimmateAPIViewSet):
    template_list = "data_explorer/provider.html"
    template_retrieve = "data_explorer/entry_detail.html"

    @classmethod
    def get_table(
        cls,
        request,
        provider_name,
        pk=None,
    ) -> DatabaseTable:
        """
        grabs the relevant database table using the URL request
        """
        # using the provider name (which is really just the table name), load
        # the corresponding database table
        #   provider_table = DatabaseTable.get_table(provider_name)
        provider_table = ALL_PROVIDERS[provider_name]
        return provider_table

    def get_list_context(
        self,
        request,
        provider_name,
    ) -> dict:
        provider_table = ALL_PROVIDERS[provider_name]
        return {
            "provider": provider_table,
            **provider_table.extra_html_context,
        }

    def get_retrieve_context(
        self,
        request,
        provider_name,
        pk,
    ) -> dict:
        return {
            "provider_name": provider_name,
            "entry_id": pk,
        }



# -*- coding: utf-8 -*-

from simmate.apps.materials_project.workflows.relaxation.mvl_grainboundary import (
    Relaxation__Vasp__MvlGrainboundary,
)


class Relaxation__Vasp__MvlSlab(Relaxation__Vasp__MvlGrainboundary):
    """
    This task is a reimplementation of pymatgen's
    [MVLGBSet](https://pymatgen.org/pymatgen.io.vasp.sets.html#pymatgen.io.vasp.sets.MVLGBSet)
    with slab_mode=True.
    """

    description_doc_short = "based on pymatgen's MVLGBSet(slab=True)"

    _incar_updates = dict(
        ISIF=2,
        NELMIN=8,
    )



# -*- coding: utf-8 -*-

from simmate.database.base_data_types import DatabaseTable, table_column


class FingerprintPool(DatabaseTable):
    """
    A table that stores Structure fingerprints

    Interaction with this table is typically done through the fingerprint
    Validator toolkit classes
    """

    class Meta:
        app_label = "core_components"
        # unique_together = ["method", "init_kwargs", "database_table"]

    method = table_column.JSONField(blank=True, null=True)
    """
    The fingerprint validator class used
    """

    init_kwargs = table_column.JSONField(blank=True, null=True)
    """
    The kwargs used to initialized the fingerprint validator class used
    """

    database_table = table_column.CharField(max_length=50, blank=True, null=True)
    """
    The database table that structures are being pulled from. 
    
    We require this to be set to a single table because the id of each 
    Fingerprint should match the id of the structure it came from. Note, 
    this is not a relation to the Structure table because its an abstract model.
    """


class Fingerprint(DatabaseTable):
    """
    A table that stores Structure fingerprints

    Interaction with this table is typically done through the fingerprint
    Validator toolkit classes
    """

    class Meta:
        app_label = "core_components"

    database_id = table_column.IntegerField(blank=True, null=True)
    """
    The id of the structure that this fingerprint came from
    """

    fingerprint = table_column.JSONField(blank=True, null=True)
    """
    The resulting fingerprint. This is typically a 1D array.
    """

    pool = table_column.ForeignKey(
        FingerprintPool,
        on_delete=table_column.CASCADE,
        related_name="fingerprints",
        blank=True,
        null=True,
    )
    """
    The database pool that this fingerprint belongs to. The pool contians useful
    information such as the fingerprint method, kwargs used, and the table that
    structures are being pulled from.
    """

    @property
    def source(self):
        # NOTE: The id of this fingerprint should match the id of the structure
        # it came from.
        return {
            "database_table": self.pool.database_table,
            "database_id": self.database_id,
        }



# -*- coding: utf-8 -*-

import logging

from simmate.apps.vasp.inputs.potcar_mappings import FOLDER_MAPPINGS
from simmate.configuration.utilities import (
    check_app_reg,
    check_command_exists,
    show_test_results,
)


def test_config(run_calcs: bool = False):
    """
    Checks to see if VASP is configured correctly
    """

    # 1 - check that VASP Simmate app is registered
    is_registered = check_app_reg("simmate.apps.configs.VaspConfig")

    # 2 - check for vasp command
    found_command = check_command_exists("vasp_std")

    # 3 - check for vasp potcars
    has_potcars = True
    for potential, folder in FOLDER_MAPPINGS.items():
        if folder.exists():
            logging.info(f"{potential} POTCARS found :heavy_check_mark:")
        else:
            logging.warning(
                f"{potential} POTCARS not found. These should be placed at... "
                f"'{folder}'"
            )
            has_potcars = False

    # 4 - run some sample workflows
    if run_calcs:
        raise NotImplementedError("This test has not been added yet.")

    # 5 - read out result of all tests
    passed_all = bool(is_registered and found_command and has_potcars)
    show_test_results("VASP", passed_all)



# -*- coding: utf-8 -*-

from simmate.apps.quantum_espresso.outputs import PwscfXml
from simmate.apps.vasp.outputs import Vasprun
from simmate.database.base_data_types import (
    Calculation,
    Forces,
    Structure,
    Thermodynamics,
    table_column,
)

# OPTIMIZE: consider lazy loading PwscfXml and Vasprun bc these apps are optional


class StaticEnergy(Structure, Thermodynamics, Forces, Calculation):
    class Meta:
        app_label = "workflows"

    archive_fields = [
        "valence_band_maximum",
        "conduction_band_minimum",
        "energy_fermi",
        "is_gap_direct",
    ]

    api_filters = dict(
        band_gap=["range"],
        is_gap_direct=["exact"],
        energy_fermi=["range"],
        conduction_band_minimum=["range"],
        valence_band_maximum=["range"],
    )

    # OPTIMIZE: should I include this electronic data?

    band_gap = table_column.FloatField(blank=True, null=True)
    """
    The band gap energy in eV.
    """

    is_gap_direct = table_column.BooleanField(blank=True, null=True)
    """
    Whether the band gap is direct or indirect.
    """

    energy_fermi = table_column.FloatField(blank=True, null=True)
    """
    The Fermi energy in eV.
    """

    conduction_band_minimum = table_column.FloatField(blank=True, null=True)
    """
    The conduction band minimum in eV.
    """

    valence_band_maximum = table_column.FloatField(blank=True, null=True)
    """
    The valence band maximum in eV.
    """

    @classmethod
    def from_vasp_run(cls, vasprun: Vasprun, as_dict: bool = False):
        # Takes a pymatgen VaspRun object, which is what's typically returned
        # from a simmate VaspWorkflow.run() call.

        # The data is actually easier to access as a dictionary and everything
        # we need is stored under the "output" key.
        data = vasprun.as_dict()["output"]
        # In a static energy calculation, there is only one ionic step so we
        # grab that up front.
        ionic_step = data["ionic_steps"][0]

        # Take our structure, energy, and forces to build all of our other
        # fields for this datatable
        static_energy = cls.from_toolkit(
            structure=vasprun.final_structure,
            energy=ionic_step["e_wo_entrp"],
            site_forces=ionic_step["forces"],
            lattice_stress=ionic_step["stress"],
            band_gap=data.get("bandgap"),
            is_gap_direct=data.get("is_gap_direct"),
            energy_fermi=data.get("efermi"),
            conduction_band_minimum=data.get("cbm"),
            valence_band_maximum=data.get("vbm"),
            as_dict=as_dict,
        )

        # If we don't want the data as a dictionary, then we are saving a new
        # object and can go ahead and do that here.
        if not as_dict:
            static_energy.save()

        return static_energy

    @classmethod
    def from_pwscf_run(cls, pwscf_run: PwscfXml, as_dict: bool = False):
        # Take our structure, energy, and forces to build all of our other
        # fields for this datatable
        static_energy = cls.from_toolkit(
            structure=pwscf_run.final_structure,
            energy=pwscf_run.final_energy,
            site_forces=pwscf_run.site_forces.tolist(),
            lattice_stress=pwscf_run.lattice_stress.tolist(),
            # TODO: I have not parsed this info out yet
            # band_gap=None,
            # is_gap_direct=None,
            # energy_fermi=None,
            # conduction_band_minimum=None,
            # valence_band_maximum=None,
            as_dict=as_dict,
        )

        # If we don't want the data as a dictionary, then we are saving a new
        # object and can go ahead and do that here.
        if not as_dict:
            static_energy.save()

        return static_energy



# -*- coding: utf-8 -*-

import pytest

from simmate.apps.materials_project.workflows.diffusion import (
    Diffusion__Vasp__NebAllPathsMit,
    Diffusion__Vasp__NebFromImagesMit,
)
from simmate.apps.vasp.inputs import PBE_POTCAR_MAPPINGS_LOW_QUALITY
from simmate.conftest import SimmateMockHelper, copy_test_files
from simmate.toolkit.diffusion import MigrationImages


@pytest.mark.slow
@pytest.mark.django_db
def test_neb_all_paths(sample_structures, tmp_path, mocker):
    SimmateMockHelper.mock_vasp(mocker)
    copy_test_files(
        tmp_path,
        test_directory=__file__,
        test_folder="nudged_elastic_band.zip",
    )

    # For testing, look at I- diffusion in Y2CF2
    structure = sample_structures["Y2CI2_mp-1206803_primitive"]

    # run the workflow and make sure it handles data properly.
    state = Diffusion__Vasp__NebAllPathsMit.run(
        structure=structure,
        migrating_specie="I",
        # command="dummycmd1; dummycmd2; dummycmd3",
        directory=tmp_path,
    )
    assert state.is_completed()

    # estabilish filenames that we make and commonly reference
    path_dir = "diffusion.vasp.neb-single-path-mit.00"
    summary_filename = tmp_path / "simmate_summary.yaml"
    plot_filename = tmp_path / path_dir / "neb_diagram.png"
    cif_filename = tmp_path / path_dir / "simmate_path_relaxed_neb.cif"

    # run the full workup
    assert summary_filename.exists()
    assert plot_filename.exists()
    assert cif_filename.exists()


def test_neb_from_images_setup(sample_structures, tmp_path, mocker):
    Potcar = SimmateMockHelper.get_mocked_potcar(mocker, tmp_path)

    # To test this task we need to create images, which we do using I diffusion
    # in Y2CI2. We use [0] to grab the shortest path.
    structure = sample_structures["Y2CI2_mp-1206803_primitive"]
    images = MigrationImages.from_structure(structure, "I")[0]

    # estabilish filenames that we make and commonly reference
    incar_filename = tmp_path / "INCAR"
    potcar_filename = tmp_path / "POTCAR"
    # These files exist within a series of directories 00, 01,..., 05
    poscar_filenames = [tmp_path / str(n).zfill(2) / "POSCAR" for n in range(5)]

    # try to make input files in the tmp_path
    Diffusion__Vasp__NebFromImagesMit.setup(
        migration_images=images,
        directory=tmp_path,
    )
    assert incar_filename.exists()

    assert all([f.exists() for f in poscar_filenames])

    assert potcar_filename.exists()
    Potcar.to_file_from_type.assert_called_with(
        structure.composition.elements,
        "PBE",
        potcar_filename,
        PBE_POTCAR_MAPPINGS_LOW_QUALITY,
    )



# -*- coding: utf-8 -*-

from simmate.database.base_data_types import Structure, table_column


class OqmdStructure(Structure):
    """
    Crystal structures from the [OQMD](http://oqmd.org/) database.

    Currently, this table only stores strucure and thermodynamic information,
    but OQDMD has much more data available via their
    [REST API](http://oqmd.org/static/docs/restful.html) and website.
    """

    class Meta:
        app_label = "data_explorer"

    archive_fields = ["formation_energy"]
    api_filters = dict(formation_energy=["range"])
    source = "OQMD"
    source_long = "The Open Quantum Materials Database"
    homepage = "https://oqmd.org/"
    source_doi = "https://doi.org/10.1007/s11837-013-0755-4"
    remote_archive_link = "https://archives.simmate.org/OqmdStructure-2023-07-21.zip"

    id = table_column.CharField(max_length=25, primary_key=True)
    """
    The id used to represent the structure (ex: "oqmd-12345")
    """

    # OQMD did not provide energy so the Thermodynamics mix-in can't be used
    formation_energy = table_column.FloatField(blank=True, null=True)
    """
    The formation energy of the structure as provided by the OQMD.
    """

    @property
    def external_link(self) -> str:
        """
        URL to this structure in the OQMD website.
        """
        # Links to the OQMD dashboard for this structure. An example is...
        #   http://oqmd.org/materials/entry/10435
        id_number = self.id.split("-")[-1]  # this removes the "oqmd-" from the id
        return f"http://oqmd.org/materials/entry/{id_number}"

    # -------------------------------------------------------------------------

    @classmethod
    def _load_all_structures_from_files(
        cls,
        base_directory: str = "oqmd",
        only_add_new_cifs: bool = True,
    ):
        """
        Only use this function if you are part of the Simmate dev team!
        Users should instead access data via the load_remote_archive method.

        This method is for pulling OQMD data into the Simmate database.

        There are many ways to pull from this database, but it looks like the easiest
        is the qmpy_rester python package. This officially supported and maintained
        at https://github.com/mohanliu/qmpy_rester. For now the package is only available
        via a pip install.

        For other options such as the REST API, check out
        http://oqmd.org/static/docs/restful.html

        # For this method specifically...

        Jiahong Shen was kind enough to provide all the crystal structures from
        the OQMD as POSCAR files. This makes loading the structures into the
        Simmate database much faster as we are no longer bottlenecked by the REST
        API and internet connections.

        All POSCARs are in the same folder, where the name of each is the
        <id>-<composition> (ex: 12345-NaCl). There are also csv's that contain
        additional data such as the energy:

            - all_oqmd_entry.csv
            - all_public_entries.csv
            - all_public_fes.csv
            - get_all_entry_id_public.py
            - get_all_entry_poscar.py

        There are currently 1,013,654 structures and this function takes roughly
        3hrs to run.
        """

        from pathlib import Path

        import pandas
        from rich.progress import track

        from simmate.toolkit import Structure as ToolkitStructure

        base_directory = Path(base_directory)

        # load the csv that contains the list of filenames and their values
        df = pandas.read_csv(base_directory / "all_oqmd_entry.csv")
        # df = df[:1000]  # FOR TESTING

        # iterate through the list and load the structures to our database!
        # Use rich to monitor progress.
        failed_entries = []
        db_objects = []
        for _, row in track(df.iterrows(), total=len(df)):
            # load the structure from the poscar file
            filename = base_directory / row.filename
            with filename.open() as file:
                contents = file.read()
            structure = ToolkitStructure.from_str(contents, "poscar")

            # save the data to the Simmate database
            # now convert the entry to a database object
            try:
                structure_db = cls.from_toolkit(
                    id="oqmd-" + str(row.entry_id),
                    structure=structure,
                    formation_energy=row.formationenergy,
                )
                db_objects.append(structure_db)
            # A few structures fail because of the symmetry analyzer can't determine
            # the spacegroup. These are...
            # 1443135, 1443014, 1451986, 1452015, 1452024
            except:
                failed_entries.append(row.entry_id)

        # and save it to our database
        cls.objects.bulk_create(
            db_objects,
            batch_size=15000,
            ignore_conflicts=True,
        )

        return failed_entries

    @classmethod
    def _load_all_structures_from_api(cls):
        """
        Only use this function if you are part of the Simmate dev team!

        Loads all structures directly for the OQMD database into the local
        Simmate database.

        THIS METHOD IS OUTDATED -- Use the _load_all_structures_from_files
        method instead
        """
        from rich.progress import track

        from simmate.toolkit import Structure as ToolkitStructure

        # OQMD is not a dependency of simmate, so make sure you install it before
        # using this method
        try:
            import qmpy_rester
        except:
            raise ModuleNotFoundError(
                "You must install qmpy-rester with `pip install qmpy_rester`"
            )
        # The documentation indicates that we handle a query via a context manager
        # for qmpy_rester. Each query is returned as a page of data, where we need
        # to iterate through all of the pages. To do this, we constantly make a query
        # and check the "next" field, which tells us if its the last page or not.

        # as we download all the data, we store it in a main list
        data = []

        # starting from the first page, assume we aren't on the last page until told
        # otherwise. And loop until we know its the last page.
        current_page = 0
        is_last_page = False
        results_per_page = 100  # based on OQMD recommendations
        while not is_last_page:
            with qmpy_rester.QMPYRester() as query:
                # make the query
                result = query.get_oqmd_phases(
                    verbose=False,
                    limit=results_per_page,
                    offset=current_page * results_per_page,
                    #
                    # Note delta_e is the formation energy and then stability is the
                    # energy above hull.
                    fields="entry_id,unit_cell,sites,delta_e",
                    # element_set="Al,C",  # Useful for testing
                )
                # grab the data for the next slice of structures
                query_slice = result["data"]
                # And check to see if this is the last page. The logic here is if
                # there is data for "next", then it isn't the last page
                is_last_page = not bool(result["links"]["next"])

                # store the slice of entrys in our main list
                for entry in query_slice:
                    data.append(entry)
                # on the very first page, let's also check the total number of pages
                if current_page == 0:
                    total_pages = result["meta"]["data_available"] // results_per_page
                # move on to the next page
                print(f"Successfully downloaded page {current_page} of {total_pages}")
                current_page += 1
        # Now iterate through all the data -- which is a list of dictionaries.
        # We convert the data into a pymatgen object and sanitize it before saving
        # to the Simmate database
        for entry in track(data):
            # Parse the data into a pymatgen object
            # Also before converting into a pymatgen object, we need to parse the sites,
            # which are given as a list of "Element @ X Y Z" (ex: "Na @ 0.5 0.5 0.5")
            # Changing this format is why we have this complex lists below
            structure = ToolkitStructure(
                lattice=entry["unit_cell"],
                species=[site.split(" @ ")[0] for site in entry["sites"]],
                coords=[
                    [float(n) for n in site.split(" @ ")[1].split()]
                    for site in entry["sites"]
                ],
                coords_are_cartesian=False,
            )

            # now convert the entry to a database object
            structure_db = cls.from_toolkit(
                id="oqmd-" + str(entry["entry_id"]),
                structure=structure,
                energy=entry["delta_e"],
            )

            # and save it to our database!
            structure_db.save()



# Generated by Django 4.2.2 on 2023-07-21 16:22

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("core_components", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="OqmdStructure",
            fields=[
                (
                    "created_at",
                    models.DateTimeField(auto_now_add=True, db_index=True, null=True),
                ),
                (
                    "updated_at",
                    models.DateTimeField(auto_now=True, db_index=True, null=True),
                ),
                ("structure", models.TextField(blank=True, null=True)),
                ("nsites", models.IntegerField(blank=True, null=True)),
                ("nelements", models.IntegerField(blank=True, null=True)),
                ("elements", models.JSONField(blank=True, null=True)),
                (
                    "chemical_system",
                    models.CharField(blank=True, max_length=25, null=True),
                ),
                ("density", models.FloatField(blank=True, null=True)),
                ("density_atomic", models.FloatField(blank=True, null=True)),
                ("volume", models.FloatField(blank=True, null=True)),
                ("volume_molar", models.FloatField(blank=True, null=True)),
                (
                    "formula_full",
                    models.CharField(blank=True, max_length=50, null=True),
                ),
                (
                    "formula_reduced",
                    models.CharField(blank=True, max_length=50, null=True),
                ),
                (
                    "formula_anonymous",
                    models.CharField(blank=True, max_length=50, null=True),
                ),
                (
                    "id",
                    models.CharField(max_length=25, primary_key=True, serialize=False),
                ),
                ("formation_energy", models.FloatField(blank=True, null=True)),
                (
                    "spacegroup",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        to="core_components.spacegroup",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="MatprojStructure",
            fields=[
                (
                    "created_at",
                    models.DateTimeField(auto_now_add=True, db_index=True, null=True),
                ),
                ("structure", models.TextField(blank=True, null=True)),
                ("nsites", models.IntegerField(blank=True, null=True)),
                ("nelements", models.IntegerField(blank=True, null=True)),
                ("elements", models.JSONField(blank=True, null=True)),
                (
                    "chemical_system",
                    models.CharField(blank=True, max_length=25, null=True),
                ),
                ("density", models.FloatField(blank=True, null=True)),
                ("density_atomic", models.FloatField(blank=True, null=True)),
                ("volume", models.FloatField(blank=True, null=True)),
                ("volume_molar", models.FloatField(blank=True, null=True)),
                (
                    "formula_full",
                    models.CharField(blank=True, max_length=50, null=True),
                ),
                (
                    "formula_reduced",
                    models.CharField(blank=True, max_length=50, null=True),
                ),
                (
                    "formula_anonymous",
                    models.CharField(blank=True, max_length=50, null=True),
                ),
                ("energy", models.FloatField(blank=True, null=True)),
                ("energy_per_atom", models.FloatField(blank=True, null=True)),
                ("energy_above_hull", models.FloatField(blank=True, null=True)),
                ("is_stable", models.BooleanField(blank=True, null=True)),
                ("decomposes_to", models.JSONField(blank=True, null=True)),
                ("formation_energy", models.FloatField(blank=True, null=True)),
                ("formation_energy_per_atom", models.FloatField(blank=True, null=True)),
                (
                    "id",
                    models.CharField(max_length=25, primary_key=True, serialize=False),
                ),
                ("energy_uncorrected", models.FloatField(blank=True, null=True)),
                ("band_gap", models.FloatField(blank=True, null=True)),
                ("is_gap_direct", models.BooleanField(blank=True, null=True)),
                ("is_magnetic", models.BooleanField(blank=True, null=True)),
                ("total_magnetization", models.FloatField(blank=True, null=True)),
                ("is_theoretical", models.BooleanField(blank=True, null=True)),
                ("updated_at", models.DateTimeField(blank=True, null=True)),
                (
                    "spacegroup",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        to="core_components.spacegroup",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="JarvisStructure",
            fields=[
                (
                    "created_at",
                    models.DateTimeField(auto_now_add=True, db_index=True, null=True),
                ),
                (
                    "updated_at",
                    models.DateTimeField(auto_now=True, db_index=True, null=True),
                ),
                ("structure", models.TextField(blank=True, null=True)),
                ("nsites", models.IntegerField(blank=True, null=True)),
                ("nelements", models.IntegerField(blank=True, null=True)),
                ("elements", models.JSONField(blank=True, null=True)),
                (
                    "chemical_system",
                    models.CharField(blank=True, max_length=25, null=True),
                ),
                ("density", models.FloatField(blank=True, null=True)),
                ("density_atomic", models.FloatField(blank=True, null=True)),
                ("volume", models.FloatField(blank=True, null=True)),
                ("volume_molar", models.FloatField(blank=True, null=True)),
                (
                    "formula_full",
                    models.CharField(blank=True, max_length=50, null=True),
                ),
                (
                    "formula_reduced",
                    models.CharField(blank=True, max_length=50, null=True),
                ),
                (
                    "formula_anonymous",
                    models.CharField(blank=True, max_length=50, null=True),
                ),
                (
                    "id",
                    models.CharField(max_length=25, primary_key=True, serialize=False),
                ),
                ("energy_above_hull", models.FloatField(blank=True, null=True)),
                (
                    "spacegroup",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        to="core_components.spacegroup",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="CodStructure",
            fields=[
                (
                    "created_at",
                    models.DateTimeField(auto_now_add=True, db_index=True, null=True),
                ),
                (
                    "updated_at",
                    models.DateTimeField(auto_now=True, db_index=True, null=True),
                ),
                ("structure", models.TextField(blank=True, null=True)),
                ("nsites", models.IntegerField(blank=True, null=True)),
                ("nelements", models.IntegerField(blank=True, null=True)),
                ("elements", models.JSONField(blank=True, null=True)),
                ("density", models.FloatField(blank=True, null=True)),
                ("density_atomic", models.FloatField(blank=True, null=True)),
                ("volume", models.FloatField(blank=True, null=True)),
                ("volume_molar", models.FloatField(blank=True, null=True)),
                ("chemical_system", models.TextField()),
                ("formula_full", models.TextField()),
                ("formula_reduced", models.TextField()),
                ("formula_anonymous", models.TextField()),
                (
                    "id",
                    models.CharField(max_length=25, primary_key=True, serialize=False),
                ),
                ("is_ordered", models.BooleanField()),
                ("has_implicit_hydrogens", models.BooleanField()),
                (
                    "spacegroup",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        to="core_components.spacegroup",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="AflowStructure",
            fields=[
                (
                    "created_at",
                    models.DateTimeField(auto_now_add=True, db_index=True, null=True),
                ),
                (
                    "updated_at",
                    models.DateTimeField(auto_now=True, db_index=True, null=True),
                ),
                ("structure", models.TextField(blank=True, null=True)),
                ("nsites", models.IntegerField(blank=True, null=True)),
                ("nelements", models.IntegerField(blank=True, null=True)),
                ("elements", models.JSONField(blank=True, null=True)),
                (
                    "chemical_system",
                    models.CharField(blank=True, max_length=25, null=True),
                ),
                ("density", models.FloatField(blank=True, null=True)),
                ("density_atomic", models.FloatField(blank=True, null=True)),
                ("volume", models.FloatField(blank=True, null=True)),
                ("volume_molar", models.FloatField(blank=True, null=True)),
                (
                    "formula_full",
                    models.CharField(blank=True, max_length=50, null=True),
                ),
                (
                    "formula_reduced",
                    models.CharField(blank=True, max_length=50, null=True),
                ),
                (
                    "formula_anonymous",
                    models.CharField(blank=True, max_length=50, null=True),
                ),
                ("energy", models.FloatField(blank=True, null=True)),
                ("energy_per_atom", models.FloatField(blank=True, null=True)),
                ("energy_above_hull", models.FloatField(blank=True, null=True)),
                ("is_stable", models.BooleanField(blank=True, null=True)),
                ("decomposes_to", models.JSONField(blank=True, null=True)),
                ("formation_energy", models.FloatField(blank=True, null=True)),
                ("formation_energy_per_atom", models.FloatField(blank=True, null=True)),
                (
                    "id",
                    models.CharField(max_length=25, primary_key=True, serialize=False),
                ),
                (
                    "spacegroup",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        to="core_components.spacegroup",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="AflowPrototype",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "created_at",
                    models.DateTimeField(auto_now_add=True, db_index=True, null=True),
                ),
                (
                    "updated_at",
                    models.DateTimeField(auto_now=True, db_index=True, null=True),
                ),
                ("structure", models.TextField(blank=True, null=True)),
                ("nsites", models.IntegerField(blank=True, null=True)),
                ("nelements", models.IntegerField(blank=True, null=True)),
                ("elements", models.JSONField(blank=True, null=True)),
                (
                    "chemical_system",
                    models.CharField(blank=True, max_length=25, null=True),
                ),
                ("density", models.FloatField(blank=True, null=True)),
                ("density_atomic", models.FloatField(blank=True, null=True)),
                ("volume", models.FloatField(blank=True, null=True)),
                ("volume_molar", models.FloatField(blank=True, null=True)),
                (
                    "formula_full",
                    models.CharField(blank=True, max_length=50, null=True),
                ),
                (
                    "formula_reduced",
                    models.CharField(blank=True, max_length=50, null=True),
                ),
                (
                    "formula_anonymous",
                    models.CharField(blank=True, max_length=50, null=True),
                ),
                (
                    "mineral_name",
                    models.CharField(blank=True, max_length=75, null=True),
                ),
                ("aflow_id", models.CharField(max_length=30)),
                ("pearson_symbol", models.CharField(max_length=6)),
                (
                    "strukturbericht_symbol",
                    models.CharField(blank=True, max_length=6, null=True),
                ),
                ("nsites_wyckoff", models.IntegerField()),
                (
                    "spacegroup",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        to="core_components.spacegroup",
                    ),
                ),
            ],
        ),
    ]



# Generated by Django 4.2.6 on 2023-11-19 21:07

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("core_components", "0002_alter_fingerprint_id_alter_fingerprintpool_id"),
    ]

    operations = [
        migrations.CreateModel(
            name="ChemicalSystemSearch",
            fields=[
                ("id", models.AutoField(primary_key=True, serialize=False)),
                (
                    "created_at",
                    models.DateTimeField(auto_now_add=True, db_index=True, null=True),
                ),
                (
                    "updated_at",
                    models.DateTimeField(auto_now=True, db_index=True, null=True),
                ),
                ("source", models.JSONField(blank=True, null=True)),
                ("started_at", models.DateTimeField(blank=True, null=True)),
                ("finished_at", models.DateTimeField(blank=True, null=True)),
                ("total_time", models.FloatField(blank=True, null=True)),
                ("queue_time", models.FloatField(blank=True, null=True)),
                (
                    "workflow_name",
                    models.CharField(blank=True, max_length=75, null=True),
                ),
                (
                    "workflow_version",
                    models.CharField(blank=True, max_length=75, null=True),
                ),
                (
                    "computer_system",
                    models.CharField(blank=True, max_length=75, null=True),
                ),
                ("directory", models.CharField(blank=True, max_length=250, null=True)),
                (
                    "run_id",
                    models.CharField(blank=True, max_length=50, null=True, unique=True),
                ),
                ("corrections", models.JSONField(blank=True, null=True)),
                (
                    "chemical_system",
                    models.CharField(blank=True, max_length=10, null=True),
                ),
                (
                    "subworkflow_name",
                    models.CharField(blank=True, max_length=200, null=True),
                ),
                (
                    "subworkflow_kwargs",
                    models.JSONField(blank=True, default=dict, null=True),
                ),
                ("max_atoms", models.IntegerField(blank=True, null=True)),
                ("max_stoich_factor", models.IntegerField(blank=True, null=True)),
                (
                    "singleshot_sources",
                    models.JSONField(blank=True, default=list, null=True),
                ),
            ],
        ),
        migrations.CreateModel(
            name="DiffusionAnalysis",
            fields=[
                ("id", models.AutoField(primary_key=True, serialize=False)),
                (
                    "created_at",
                    models.DateTimeField(auto_now_add=True, db_index=True, null=True),
                ),
                (
                    "updated_at",
                    models.DateTimeField(auto_now=True, db_index=True, null=True),
                ),
                ("source", models.JSONField(blank=True, null=True)),
                ("structure", models.TextField(blank=True, null=True)),
                ("nsites", models.IntegerField(blank=True, null=True)),
                ("nelements", models.IntegerField(blank=True, null=True)),
                ("elements", models.JSONField(blank=True, null=True)),
                (
                    "chemical_system",
                    models.CharField(blank=True, max_length=25, null=True),
                ),
                ("density", models.FloatField(blank=True, null=True)),
                ("density_atomic", models.FloatField(blank=True, null=True)),
                ("volume", models.FloatField(blank=True, null=True)),
                ("volume_molar", models.FloatField(blank=True, null=True)),
                (
                    "formula_full",
                    models.CharField(blank=True, max_length=50, null=True),
                ),
                (
                    "formula_reduced",
                    models.CharField(blank=True, max_length=50, null=True),
                ),
                (
                    "formula_anonymous",
                    models.CharField(blank=True, max_length=50, null=True),
                ),
                ("started_at", models.DateTimeField(blank=True, null=True)),
                ("finished_at", models.DateTimeField(blank=True, null=True)),
                ("total_time", models.FloatField(blank=True, null=True)),
                ("queue_time", models.FloatField(blank=True, null=True)),
                (
                    "workflow_name",
                    models.CharField(blank=True, max_length=75, null=True),
                ),
                (
                    "workflow_version",
                    models.CharField(blank=True, max_length=75, null=True),
                ),
                (
                    "computer_system",
                    models.CharField(blank=True, max_length=75, null=True),
                ),
                ("directory", models.CharField(blank=True, max_length=250, null=True)),
                (
                    "run_id",
                    models.CharField(blank=True, max_length=50, null=True, unique=True),
                ),
                ("corrections", models.JSONField(blank=True, null=True)),
                (
                    "migrating_specie",
                    models.CharField(blank=True, max_length=4, null=True),
                ),
                ("vacancy_mode", models.BooleanField(blank=True, null=True)),
                ("atomic_fraction", models.FloatField(blank=True, null=True)),
                ("barrier_cell", models.FloatField(blank=True, null=True)),
                (
                    "paths_involved",
                    models.CharField(blank=True, max_length=100, null=True),
                ),
                ("npaths_involved", models.IntegerField(blank=True, null=True)),
                (
                    "spacegroup",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        to="core_components.spacegroup",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Dynamics",
            fields=[
                ("id", models.AutoField(primary_key=True, serialize=False)),
                (
                    "created_at",
                    models.DateTimeField(auto_now_add=True, db_index=True, null=True),
                ),
                (
                    "updated_at",
                    models.DateTimeField(auto_now=True, db_index=True, null=True),
                ),
                ("source", models.JSONField(blank=True, null=True)),
                ("structure", models.TextField(blank=True, null=True)),
                ("nsites", models.IntegerField(blank=True, null=True)),
                ("nelements", models.IntegerField(blank=True, null=True)),
                ("elements", models.JSONField(blank=True, null=True)),
                (
                    "chemical_system",
                    models.CharField(blank=True, max_length=25, null=True),
                ),
                ("density", models.FloatField(blank=True, null=True)),
                ("density_atomic", models.FloatField(blank=True, null=True)),
                ("volume", models.FloatField(blank=True, null=True)),
                ("volume_molar", models.FloatField(blank=True, null=True)),
                (
                    "formula_full",
                    models.CharField(blank=True, max_length=50, null=True),
                ),
                (
                    "formula_reduced",
                    models.CharField(blank=True, max_length=50, null=True),
                ),
                (
                    "formula_anonymous",
                    models.CharField(blank=True, max_length=50, null=True),
                ),
                ("started_at", models.DateTimeField(blank=True, null=True)),
                ("finished_at", models.DateTimeField(blank=True, null=True)),
                ("total_time", models.FloatField(blank=True, null=True)),
                ("queue_time", models.FloatField(blank=True, null=True)),
                (
                    "workflow_name",
                    models.CharField(blank=True, max_length=75, null=True),
                ),
                (
                    "workflow_version",
                    models.CharField(blank=True, max_length=75, null=True),
                ),
                (
                    "computer_system",
                    models.CharField(blank=True, max_length=75, null=True),
                ),
                ("directory", models.CharField(blank=True, max_length=250, null=True)),
                (
                    "run_id",
                    models.CharField(blank=True, max_length=50, null=True, unique=True),
                ),
                ("corrections", models.JSONField(blank=True, null=True)),
                ("temperature_start", models.IntegerField(blank=True, null=True)),
                ("temperature_end", models.IntegerField(blank=True, null=True)),
                ("time_step", models.FloatField(blank=True, null=True)),
                ("nsteps", models.IntegerField(blank=True, null=True)),
                (
                    "spacegroup",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        to="core_components.spacegroup",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="FixedCompositionSearch",
            fields=[
                ("id", models.AutoField(primary_key=True, serialize=False)),
                (
                    "created_at",
                    models.DateTimeField(auto_now_add=True, db_index=True, null=True),
                ),
                (
                    "updated_at",
                    models.DateTimeField(auto_now=True, db_index=True, null=True),
                ),
                ("source", models.JSONField(blank=True, null=True)),
                ("started_at", models.DateTimeField(blank=True, null=True)),
                ("finished_at", models.DateTimeField(blank=True, null=True)),
                ("total_time", models.FloatField(blank=True, null=True)),
                ("queue_time", models.FloatField(blank=True, null=True)),
                (
                    "workflow_name",
                    models.CharField(blank=True, max_length=75, null=True),
                ),
                (
                    "workflow_version",
                    models.CharField(blank=True, max_length=75, null=True),
                ),
                (
                    "computer_system",
                    models.CharField(blank=True, max_length=75, null=True),
                ),
                ("directory", models.CharField(blank=True, max_length=250, null=True)),
                (
                    "run_id",
                    models.CharField(blank=True, max_length=50, null=True, unique=True),
                ),
                ("corrections", models.JSONField(blank=True, null=True)),
                ("composition", models.CharField(blank=True, max_length=50, null=True)),
                (
                    "subworkflow_name",
                    models.CharField(blank=True, max_length=200, null=True),
                ),
                (
                    "subworkflow_kwargs",
                    models.JSONField(blank=True, default=dict, null=True),
                ),
                (
                    "fitness_field",
                    models.CharField(blank=True, max_length=200, null=True),
                ),
                ("min_structures_exact", models.IntegerField(blank=True, null=True)),
                ("max_structures", models.IntegerField(blank=True, null=True)),
                ("best_survival_cutoff", models.IntegerField(blank=True, null=True)),
                ("nfirst_generation", models.IntegerField(blank=True, null=True)),
                ("nsteadystate", models.IntegerField(blank=True, null=True)),
                ("convergence_cutoff", models.FloatField(blank=True, null=True)),
                (
                    "selector_name",
                    models.CharField(blank=True, max_length=200, null=True),
                ),
                (
                    "selector_kwargs",
                    models.JSONField(blank=True, default=dict, null=True),
                ),
                (
                    "validator_name",
                    models.CharField(blank=True, max_length=200, null=True),
                ),
                (
                    "validator_kwargs",
                    models.JSONField(blank=True, default=dict, null=True),
                ),
                (
                    "singleshot_sources",
                    models.JSONField(blank=True, default=list, null=True),
                ),
                ("sleep_step", models.FloatField(blank=True, null=True)),
                (
                    "unique_individuals_ids",
                    models.JSONField(blank=True, default=list, null=True),
                ),
                ("expected_structure", models.JSONField(default=dict)),
            ],
        ),
        migrations.CreateModel(
            name="IonicStep",
            fields=[
                ("id", models.AutoField(primary_key=True, serialize=False)),
                (
                    "created_at",
                    models.DateTimeField(auto_now_add=True, db_index=True, null=True),
                ),
                (
                    "updated_at",
                    models.DateTimeField(auto_now=True, db_index=True, null=True),
                ),
                ("source", models.JSONField(blank=True, null=True)),
                ("structure", models.TextField(blank=True, null=True)),
                ("nsites", models.IntegerField(blank=True, null=True)),
                ("nelements", models.IntegerField(blank=True, null=True)),
                ("elements", models.JSONField(blank=True, null=True)),
                (
                    "chemical_system",
                    models.CharField(blank=True, max_length=25, null=True),
                ),
                ("density", models.FloatField(blank=True, null=True)),
                ("density_atomic", models.FloatField(blank=True, null=True)),
                ("volume", models.FloatField(blank=True, null=True)),
                ("volume_molar", models.FloatField(blank=True, null=True)),
                (
                    "formula_full",
                    models.CharField(blank=True, max_length=50, null=True),
                ),
                (
                    "formula_reduced",
                    models.CharField(blank=True, max_length=50, null=True),
                ),
                (
                    "formula_anonymous",
                    models.CharField(blank=True, max_length=50, null=True),
                ),
                ("site_forces", models.JSONField(blank=True, null=True)),
                ("lattice_stress", models.JSONField(blank=True, null=True)),
                ("site_force_norm_max", models.FloatField(blank=True, null=True)),
                ("site_forces_norm", models.FloatField(blank=True, null=True)),
                ("site_forces_norm_per_atom", models.FloatField(blank=True, null=True)),
                ("lattice_stress_norm", models.FloatField(blank=True, null=True)),
                (
                    "lattice_stress_norm_per_atom",
                    models.FloatField(blank=True, null=True),
                ),
                ("energy", models.FloatField(blank=True, null=True)),
                ("energy_per_atom", models.FloatField(blank=True, null=True)),
                ("energy_above_hull", models.FloatField(blank=True, null=True)),
                ("is_stable", models.BooleanField(blank=True, null=True)),
                ("decomposes_to", models.JSONField(blank=True, null=True)),
                ("formation_energy", models.FloatField(blank=True, null=True)),
                ("formation_energy_per_atom", models.FloatField(blank=True, null=True)),
                ("number", models.IntegerField()),
            ],
        ),
        migrations.CreateModel(
            name="MigrationHop",
            fields=[
                ("id", models.AutoField(primary_key=True, serialize=False)),
                (
                    "created_at",
                    models.DateTimeField(auto_now_add=True, db_index=True, null=True),
                ),
                (
                    "updated_at",
                    models.DateTimeField(auto_now=True, db_index=True, null=True),
                ),
                ("source", models.JSONField(blank=True, null=True)),
                ("started_at", models.DateTimeField(blank=True, null=True)),
                ("finished_at", models.DateTimeField(blank=True, null=True)),
                ("total_time", models.FloatField(blank=True, null=True)),
                ("queue_time", models.FloatField(blank=True, null=True)),
                (
                    "workflow_name",
                    models.CharField(blank=True, max_length=75, null=True),
                ),
                (
                    "workflow_version",
                    models.CharField(blank=True, max_length=75, null=True),
                ),
                (
                    "computer_system",
                    models.CharField(blank=True, max_length=75, null=True),
                ),
                ("directory", models.CharField(blank=True, max_length=250, null=True)),
                (
                    "run_id",
                    models.CharField(blank=True, max_length=50, null=True, unique=True),
                ),
                ("corrections", models.JSONField(blank=True, null=True)),
                ("site_start", models.CharField(blank=True, max_length=100, null=True)),
                ("site_end", models.CharField(blank=True, max_length=100, null=True)),
                ("index_start", models.IntegerField(blank=True, null=True)),
                ("index_end", models.IntegerField(blank=True, null=True)),
                ("number", models.IntegerField(blank=True, null=True)),
                ("length", models.FloatField(blank=True, null=True)),
                ("dimension_path", models.IntegerField(blank=True, null=True)),
                ("dimension_host_lattice", models.IntegerField(blank=True, null=True)),
                ("energy_barrier", models.FloatField(blank=True, null=True)),
                (
                    "diffusion_analysis",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="migration_hops",
                        to="workflows.diffusionanalysis",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="StaticEnergy",
            fields=[
                ("id", models.AutoField(primary_key=True, serialize=False)),
                (
                    "created_at",
                    models.DateTimeField(auto_now_add=True, db_index=True, null=True),
                ),
                (
                    "updated_at",
                    models.DateTimeField(auto_now=True, db_index=True, null=True),
                ),
                ("source", models.JSONField(blank=True, null=True)),
                ("structure", models.TextField(blank=True, null=True)),
                ("nsites", models.IntegerField(blank=True, null=True)),
                ("nelements", models.IntegerField(blank=True, null=True)),
                ("elements", models.JSONField(blank=True, null=True)),
                (
                    "chemical_system",
                    models.CharField(blank=True, max_length=25, null=True),
                ),
                ("density", models.FloatField(blank=True, null=True)),
                ("density_atomic", models.FloatField(blank=True, null=True)),
                ("volume", models.FloatField(blank=True, null=True)),
                ("volume_molar", models.FloatField(blank=True, null=True)),
                (
                    "formula_full",
                    models.CharField(blank=True, max_length=50, null=True),
                ),
                (
                    "formula_reduced",
                    models.CharField(blank=True, max_length=50, null=True),
                ),
                (
                    "formula_anonymous",
                    models.CharField(blank=True, max_length=50, null=True),
                ),
                ("started_at", models.DateTimeField(blank=True, null=True)),
                ("finished_at", models.DateTimeField(blank=True, null=True)),
                ("total_time", models.FloatField(blank=True, null=True)),
                ("queue_time", models.FloatField(blank=True, null=True)),
                (
                    "workflow_name",
                    models.CharField(blank=True, max_length=75, null=True),
                ),
                (
                    "workflow_version",
                    models.CharField(blank=True, max_length=75, null=True),
                ),
                (
                    "computer_system",
                    models.CharField(blank=True, max_length=75, null=True),
                ),
                ("directory", models.CharField(blank=True, max_length=250, null=True)),
                (
                    "run_id",
                    models.CharField(blank=True, max_length=50, null=True, unique=True),
                ),
                ("corrections", models.JSONField(blank=True, null=True)),
                ("site_forces", models.JSONField(blank=True, null=True)),
                ("lattice_stress", models.JSONField(blank=True, null=True)),
                ("site_force_norm_max", models.FloatField(blank=True, null=True)),
                ("site_forces_norm", models.FloatField(blank=True, null=True)),
                ("site_forces_norm_per_atom", models.FloatField(blank=True, null=True)),
                ("lattice_stress_norm", models.FloatField(blank=True, null=True)),
                (
                    "lattice_stress_norm_per_atom",
                    models.FloatField(blank=True, null=True),
                ),
                ("energy", models.FloatField(blank=True, null=True)),
                ("energy_per_atom", models.FloatField(blank=True, null=True)),
                ("energy_above_hull", models.FloatField(blank=True, null=True)),
                ("is_stable", models.BooleanField(blank=True, null=True)),
                ("decomposes_to", models.JSONField(blank=True, null=True)),
                ("formation_energy", models.FloatField(blank=True, null=True)),
                ("formation_energy_per_atom", models.FloatField(blank=True, null=True)),
                ("band_gap", models.FloatField(blank=True, null=True)),
                ("is_gap_direct", models.BooleanField(blank=True, null=True)),
                ("energy_fermi", models.FloatField(blank=True, null=True)),
                ("conduction_band_minimum", models.FloatField(blank=True, null=True)),
                ("valence_band_maximum", models.FloatField(blank=True, null=True)),
                (
                    "spacegroup",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        to="core_components.spacegroup",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="VariableNsitesCompositionSearch",
            fields=[
                ("id", models.AutoField(primary_key=True, serialize=False)),
                (
                    "created_at",
                    models.DateTimeField(auto_now_add=True, db_index=True, null=True),
                ),
                (
                    "updated_at",
                    models.DateTimeField(auto_now=True, db_index=True, null=True),
                ),
                ("source", models.JSONField(blank=True, null=True)),
                ("started_at", models.DateTimeField(blank=True, null=True)),
                ("finished_at", models.DateTimeField(blank=True, null=True)),
                ("total_time", models.FloatField(blank=True, null=True)),
                ("queue_time", models.FloatField(blank=True, null=True)),
                (
                    "workflow_name",
                    models.CharField(blank=True, max_length=75, null=True),
                ),
                (
                    "workflow_version",
                    models.CharField(blank=True, max_length=75, null=True),
                ),
                (
                    "computer_system",
                    models.CharField(blank=True, max_length=75, null=True),
                ),
                ("directory", models.CharField(blank=True, max_length=250, null=True)),
                (
                    "run_id",
                    models.CharField(blank=True, max_length=50, null=True, unique=True),
                ),
                ("corrections", models.JSONField(blank=True, null=True)),
                ("composition", models.CharField(blank=True, max_length=50, null=True)),
                (
                    "subworkflow_name",
                    models.CharField(blank=True, max_length=200, null=True),
                ),
                (
                    "subworkflow_kwargs",
                    models.JSONField(blank=True, default=dict, null=True),
                ),
                (
                    "fitness_field",
                    models.CharField(blank=True, max_length=200, null=True),
                ),
                ("min_structures_exact", models.IntegerField(blank=True, null=True)),
                ("max_structures", models.IntegerField(blank=True, null=True)),
                ("best_survival_cutoff", models.IntegerField(blank=True, null=True)),
                ("nfirst_generation", models.IntegerField(blank=True, null=True)),
                ("nsteadystate", models.IntegerField(blank=True, null=True)),
                ("convergence_cutoff", models.FloatField(blank=True, null=True)),
                (
                    "selector_name",
                    models.CharField(blank=True, max_length=200, null=True),
                ),
                (
                    "selector_kwargs",
                    models.JSONField(blank=True, default=dict, null=True),
                ),
                (
                    "validator_name",
                    models.CharField(blank=True, max_length=200, null=True),
                ),
                (
                    "validator_kwargs",
                    models.JSONField(blank=True, default=dict, null=True),
                ),
                ("sleep_step", models.FloatField(blank=True, null=True)),
                ("expected_structure", models.JSONField(default=dict)),
            ],
        ),
        migrations.CreateModel(
            name="PopulationAnalysis",
            fields=[
                (
                    "staticenergy_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="workflows.staticenergy",
                    ),
                ),
                ("oxidation_states", models.JSONField(blank=True, null=True)),
                ("charges", models.JSONField(blank=True, null=True)),
                ("min_dists", models.JSONField(blank=True, null=True)),
                ("atomic_volumes", models.JSONField(blank=True, null=True)),
                ("element_list", models.JSONField(blank=True, null=True)),
                ("vacuum_charge", models.FloatField(blank=True, null=True)),
                ("vacuum_volume", models.FloatField(blank=True, null=True)),
                ("nelectrons", models.FloatField(blank=True, null=True)),
            ],
            bases=("workflows.staticenergy",),
        ),
        migrations.CreateModel(
            name="SteadystateSource",
            fields=[
                ("id", models.AutoField(primary_key=True, serialize=False)),
                (
                    "created_at",
                    models.DateTimeField(auto_now_add=True, db_index=True, null=True),
                ),
                (
                    "updated_at",
                    models.DateTimeField(auto_now=True, db_index=True, null=True),
                ),
                ("source", models.JSONField(blank=True, null=True)),
                ("name", models.CharField(max_length=50)),
                ("kwargs", models.JSONField(default=dict)),
                ("nsteadystate_target", models.FloatField(blank=True, null=True)),
                ("is_creator", models.BooleanField()),
                ("is_transformation", models.BooleanField()),
                ("workitem_ids", models.JSONField(default=list)),
                (
                    "search",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="steadystate_sources",
                        to="workflows.fixedcompositionsearch",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Relaxation",
            fields=[
                ("id", models.AutoField(primary_key=True, serialize=False)),
                (
                    "created_at",
                    models.DateTimeField(auto_now_add=True, db_index=True, null=True),
                ),
                (
                    "updated_at",
                    models.DateTimeField(auto_now=True, db_index=True, null=True),
                ),
                ("source", models.JSONField(blank=True, null=True)),
                ("structure", models.TextField(blank=True, null=True)),
                ("nsites", models.IntegerField(blank=True, null=True)),
                ("nelements", models.IntegerField(blank=True, null=True)),
                ("elements", models.JSONField(blank=True, null=True)),
                (
                    "chemical_system",
                    models.CharField(blank=True, max_length=25, null=True),
                ),
                ("density", models.FloatField(blank=True, null=True)),
                ("density_atomic", models.FloatField(blank=True, null=True)),
                ("volume", models.FloatField(blank=True, null=True)),
                ("volume_molar", models.FloatField(blank=True, null=True)),
                (
                    "formula_full",
                    models.CharField(blank=True, max_length=50, null=True),
                ),
                (
                    "formula_reduced",
                    models.CharField(blank=True, max_length=50, null=True),
                ),
                (
                    "formula_anonymous",
                    models.CharField(blank=True, max_length=50, null=True),
                ),
                ("started_at", models.DateTimeField(blank=True, null=True)),
                ("finished_at", models.DateTimeField(blank=True, null=True)),
                ("total_time", models.FloatField(blank=True, null=True)),
                ("queue_time", models.FloatField(blank=True, null=True)),
                (
                    "workflow_name",
                    models.CharField(blank=True, max_length=75, null=True),
                ),
                (
                    "workflow_version",
                    models.CharField(blank=True, max_length=75, null=True),
                ),
                (
                    "computer_system",
                    models.CharField(blank=True, max_length=75, null=True),
                ),
                ("directory", models.CharField(blank=True, max_length=250, null=True)),
                (
                    "run_id",
                    models.CharField(blank=True, max_length=50, null=True, unique=True),
                ),
                ("corrections", models.JSONField(blank=True, null=True)),
                ("site_forces", models.JSONField(blank=True, null=True)),
                ("lattice_stress", models.JSONField(blank=True, null=True)),
                ("site_force_norm_max", models.FloatField(blank=True, null=True)),
                ("site_forces_norm", models.FloatField(blank=True, null=True)),
                ("site_forces_norm_per_atom", models.FloatField(blank=True, null=True)),
                ("lattice_stress_norm", models.FloatField(blank=True, null=True)),
                (
                    "lattice_stress_norm_per_atom",
                    models.FloatField(blank=True, null=True),
                ),
                ("energy", models.FloatField(blank=True, null=True)),
                ("energy_per_atom", models.FloatField(blank=True, null=True)),
                ("energy_above_hull", models.FloatField(blank=True, null=True)),
                ("is_stable", models.BooleanField(blank=True, null=True)),
                ("decomposes_to", models.JSONField(blank=True, null=True)),
                ("formation_energy", models.FloatField(blank=True, null=True)),
                ("formation_energy_per_atom", models.FloatField(blank=True, null=True)),
                ("band_gap", models.FloatField(blank=True, null=True)),
                ("is_gap_direct", models.BooleanField(blank=True, null=True)),
                ("energy_fermi", models.FloatField(blank=True, null=True)),
                ("conduction_band_minimum", models.FloatField(blank=True, null=True)),
                ("valence_band_maximum", models.FloatField(blank=True, null=True)),
                ("volume_change", models.FloatField(blank=True, null=True)),
                (
                    "spacegroup",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        to="core_components.spacegroup",
                    ),
                ),
                (
                    "structure_final",
                    models.OneToOneField(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="relaxations_as_final",
                        to="workflows.ionicstep",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="MigrationImage",
            fields=[
                ("id", models.AutoField(primary_key=True, serialize=False)),
                (
                    "created_at",
                    models.DateTimeField(auto_now_add=True, db_index=True, null=True),
                ),
                (
                    "updated_at",
                    models.DateTimeField(auto_now=True, db_index=True, null=True),
                ),
                ("structure", models.TextField(blank=True, null=True)),
                ("nsites", models.IntegerField(blank=True, null=True)),
                ("nelements", models.IntegerField(blank=True, null=True)),
                ("elements", models.JSONField(blank=True, null=True)),
                (
                    "chemical_system",
                    models.CharField(blank=True, max_length=25, null=True),
                ),
                ("density", models.FloatField(blank=True, null=True)),
                ("density_atomic", models.FloatField(blank=True, null=True)),
                ("volume", models.FloatField(blank=True, null=True)),
                ("volume_molar", models.FloatField(blank=True, null=True)),
                (
                    "formula_full",
                    models.CharField(blank=True, max_length=50, null=True),
                ),
                (
                    "formula_reduced",
                    models.CharField(blank=True, max_length=50, null=True),
                ),
                (
                    "formula_anonymous",
                    models.CharField(blank=True, max_length=50, null=True),
                ),
                ("number", models.IntegerField()),
                ("force_tangent", models.FloatField(blank=True, null=True)),
                ("energy", models.FloatField(blank=True, null=True)),
                ("structure_distance", models.FloatField(blank=True, null=True)),
                (
                    "migration_hop",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="migration_images",
                        to="workflows.migrationhop",
                    ),
                ),
                (
                    "spacegroup",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        to="core_components.spacegroup",
                    ),
                ),
            ],
        ),
        migrations.AddField(
            model_name="ionicstep",
            name="relaxation",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="structures",
                to="workflows.relaxation",
            ),
        ),
        migrations.AddField(
            model_name="ionicstep",
            name="spacegroup",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to="core_components.spacegroup",
            ),
        ),
        migrations.CreateModel(
            name="DynamicsIonicStep",
            fields=[
                ("id", models.AutoField(primary_key=True, serialize=False)),
                (
                    "created_at",
                    models.DateTimeField(auto_now_add=True, db_index=True, null=True),
                ),
                (
                    "updated_at",
                    models.DateTimeField(auto_now=True, db_index=True, null=True),
                ),
                ("source", models.JSONField(blank=True, null=True)),
                ("structure", models.TextField(blank=True, null=True)),
                ("nsites", models.IntegerField(blank=True, null=True)),
                ("nelements", models.IntegerField(blank=True, null=True)),
                ("elements", models.JSONField(blank=True, null=True)),
                (
                    "chemical_system",
                    models.CharField(blank=True, max_length=25, null=True),
                ),
                ("density", models.FloatField(blank=True, null=True)),
                ("density_atomic", models.FloatField(blank=True, null=True)),
                ("volume", models.FloatField(blank=True, null=True)),
                ("volume_molar", models.FloatField(blank=True, null=True)),
                (
                    "formula_full",
                    models.CharField(blank=True, max_length=50, null=True),
                ),
                (
                    "formula_reduced",
                    models.CharField(blank=True, max_length=50, null=True),
                ),
                (
                    "formula_anonymous",
                    models.CharField(blank=True, max_length=50, null=True),
                ),
                ("site_forces", models.JSONField(blank=True, null=True)),
                ("lattice_stress", models.JSONField(blank=True, null=True)),
                ("site_force_norm_max", models.FloatField(blank=True, null=True)),
                ("site_forces_norm", models.FloatField(blank=True, null=True)),
                ("site_forces_norm_per_atom", models.FloatField(blank=True, null=True)),
                ("lattice_stress_norm", models.FloatField(blank=True, null=True)),
                (
                    "lattice_stress_norm_per_atom",
                    models.FloatField(blank=True, null=True),
                ),
                ("energy", models.FloatField(blank=True, null=True)),
                ("energy_per_atom", models.FloatField(blank=True, null=True)),
                ("energy_above_hull", models.FloatField(blank=True, null=True)),
                ("is_stable", models.BooleanField(blank=True, null=True)),
                ("decomposes_to", models.JSONField(blank=True, null=True)),
                ("formation_energy", models.FloatField(blank=True, null=True)),
                ("formation_energy_per_atom", models.FloatField(blank=True, null=True)),
                ("number", models.IntegerField()),
                ("temperature", models.FloatField(blank=True, null=True)),
                (
                    "dynamics_run",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="structures",
                        to="workflows.dynamics",
                    ),
                ),
                (
                    "spacegroup",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        to="core_components.spacegroup",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="DensityofStatesCalc",
            fields=[
                ("id", models.AutoField(primary_key=True, serialize=False)),
                (
                    "created_at",
                    models.DateTimeField(auto_now_add=True, db_index=True, null=True),
                ),
                (
                    "updated_at",
                    models.DateTimeField(auto_now=True, db_index=True, null=True),
                ),
                ("source", models.JSONField(blank=True, null=True)),
                ("structure", models.TextField(blank=True, null=True)),
                ("nsites", models.IntegerField(blank=True, null=True)),
                ("nelements", models.IntegerField(blank=True, null=True)),
                ("elements", models.JSONField(blank=True, null=True)),
                (
                    "chemical_system",
                    models.CharField(blank=True, max_length=25, null=True),
                ),
                ("density", models.FloatField(blank=True, null=True)),
                ("density_atomic", models.FloatField(blank=True, null=True)),
                ("volume", models.FloatField(blank=True, null=True)),
                ("volume_molar", models.FloatField(blank=True, null=True)),
                (
                    "formula_full",
                    models.CharField(blank=True, max_length=50, null=True),
                ),
                (
                    "formula_reduced",
                    models.CharField(blank=True, max_length=50, null=True),
                ),
                (
                    "formula_anonymous",
                    models.CharField(blank=True, max_length=50, null=True),
                ),
                ("started_at", models.DateTimeField(blank=True, null=True)),
                ("finished_at", models.DateTimeField(blank=True, null=True)),
                ("total_time", models.FloatField(blank=True, null=True)),
                ("queue_time", models.FloatField(blank=True, null=True)),
                (
                    "workflow_name",
                    models.CharField(blank=True, max_length=75, null=True),
                ),
                (
                    "workflow_version",
                    models.CharField(blank=True, max_length=75, null=True),
                ),
                (
                    "computer_system",
                    models.CharField(blank=True, max_length=75, null=True),
                ),
                ("directory", models.CharField(blank=True, max_length=250, null=True)),
                (
                    "run_id",
                    models.CharField(blank=True, max_length=50, null=True, unique=True),
                ),
                ("corrections", models.JSONField(blank=True, null=True)),
                ("density_of_states_data", models.JSONField(blank=True, null=True)),
                ("band_gap", models.FloatField(blank=True, null=True)),
                ("energy_fermi", models.FloatField(blank=True, null=True)),
                ("conduction_band_minimum", models.FloatField(blank=True, null=True)),
                ("valence_band_maximum", models.FloatField(blank=True, null=True)),
                (
                    "spacegroup",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        to="core_components.spacegroup",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="BandStructureCalc",
            fields=[
                ("id", models.AutoField(primary_key=True, serialize=False)),
                (
                    "created_at",
                    models.DateTimeField(auto_now_add=True, db_index=True, null=True),
                ),
                (
                    "updated_at",
                    models.DateTimeField(auto_now=True, db_index=True, null=True),
                ),
                ("source", models.JSONField(blank=True, null=True)),
                ("structure", models.TextField(blank=True, null=True)),
                ("nsites", models.IntegerField(blank=True, null=True)),
                ("nelements", models.IntegerField(blank=True, null=True)),
                ("elements", models.JSONField(blank=True, null=True)),
                (
                    "chemical_system",
                    models.CharField(blank=True, max_length=25, null=True),
                ),
                ("density", models.FloatField(blank=True, null=True)),
                ("density_atomic", models.FloatField(blank=True, null=True)),
                ("volume", models.FloatField(blank=True, null=True)),
                ("volume_molar", models.FloatField(blank=True, null=True)),
                (
                    "formula_full",
                    models.CharField(blank=True, max_length=50, null=True),
                ),
                (
                    "formula_reduced",
                    models.CharField(blank=True, max_length=50, null=True),
                ),
                (
                    "formula_anonymous",
                    models.CharField(blank=True, max_length=50, null=True),
                ),
                ("started_at", models.DateTimeField(blank=True, null=True)),
                ("finished_at", models.DateTimeField(blank=True, null=True)),
                ("total_time", models.FloatField(blank=True, null=True)),
                ("queue_time", models.FloatField(blank=True, null=True)),
                (
                    "workflow_name",
                    models.CharField(blank=True, max_length=75, null=True),
                ),
                (
                    "workflow_version",
                    models.CharField(blank=True, max_length=75, null=True),
                ),
                (
                    "computer_system",
                    models.CharField(blank=True, max_length=75, null=True),
                ),
                ("directory", models.CharField(blank=True, max_length=250, null=True)),
                (
                    "run_id",
                    models.CharField(blank=True, max_length=50, null=True, unique=True),
                ),
                ("corrections", models.JSONField(blank=True, null=True)),
                ("band_structure_data", models.JSONField(blank=True, null=True)),
                ("nbands", models.IntegerField(blank=True, null=True)),
                ("band_gap", models.FloatField(blank=True, null=True)),
                ("is_gap_direct", models.BooleanField(blank=True, null=True)),
                ("band_gap_direct", models.FloatField(blank=True, null=True)),
                ("energy_fermi", models.FloatField(blank=True, null=True)),
                ("conduction_band_minimum", models.FloatField(blank=True, null=True)),
                ("valence_band_maximum", models.FloatField(blank=True, null=True)),
                ("is_metal", models.BooleanField(blank=True, null=True)),
                (
                    "spacegroup",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        to="core_components.spacegroup",
                    ),
                ),
            ],
        ),
    ]



# -*- coding: utf-8 -*-

import os
import shutil
from pathlib import Path

from simmate.engine import Workflow
from simmate.toolkit import Structure


class RelaxationStaticBase(Workflow):
    """
    Base class for running a relaxation followed by a static energy
    calculation.

    This should NOT be run on its own. It is meant to be inherited from in
    other workflows.
    """

    # We don't want to save anything from the parent workflow, only the
    # sub workflows (relaxation and static energy) so we set use_database=False
    use_database = False
    relaxation_workflow = None  # This will be defined in inheriting workflows
    static_energy_workflow = None  # This will be defined in inheriting workflows

    @classmethod
    def run_config(
        cls,
        structure: Structure,
        command: str = None,
        source: dict = None,
        directory: Path = None,
        **kwargs,
    ):
        # run a relaxation
        relaxation_directory = directory / "relaxation"
        relaxation_result = cls.relaxation_workflow.run(
            structure=structure,
            command=command,
            source=source,
            directory=relaxation_directory,
        ).result()

        static_energy_directory = directory / "static_energy"
        # run a static energy and bader/badelf analysis using the same structure
        # as above.
        # !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        # We need to make a new directory because only one vasp workflow can
        # be run in each directory.
        os.mkdir(static_energy_directory)
        shutil.copyfile(
            relaxation_directory / "WAVECAR", static_energy_directory / "WAVECAR"
        )
        static_energy_result = cls.static_energy_workflow.run(
            structure=relaxation_result,
            command=command,
            source=source,
            directory=static_energy_directory,
            # copy_previous_directory=True,
        )


class RelaxationRelaxationStaticBase(Workflow):
    """
    Base class for running a PBE relaxation followed by an HSE relaxation
    and then a static energy calculation.

    This should NOT be run on its own. It is meant to be inherited from in
    other workflows.
    """

    # We don't want to save anything from the parent workflow, only the
    # sub workflows (relaxation and static energy) so we set use_database=False
    use_database = False
    low_quality_relaxation_workflow = (
        None  # This will be defined in inheriting workflows
    )
    high_quality_relaxation_workflow = (
        None  # This will be defined in inheriting workflows
    )
    static_energy_workflow = None  # This will be defined in inheriting workflows

    @classmethod
    def run_config(
        cls,
        structure: Structure,
        command: str = None,
        source: dict = None,
        directory: Path = None,
        **kwargs,
    ):
        # run a relaxation at low quality
        relaxation1_directory = directory / "low_quality_relaxation"
        relaxation1_result = cls.low_quality_relaxation_workflow.run(
            structure=structure,
            command=command,
            source=source,
            directory=relaxation1_directory,
        ).result()

        # make the directory for the second relaxation and copy the WAVECAR
        # then run
        relaxation2_directory = directory / "high_quality_relaxation"
        os.mkdir(relaxation2_directory)
        shutil.copyfile(
            relaxation1_directory / "WAVECAR", relaxation2_directory / "WAVECAR"
        )
        relaxation2_result = cls.high_quality_relaxation_workflow.run(
            structure=relaxation1_result,
            command=command,
            directory=relaxation2_directory,
            # copy_previous_directory=True
        ).result()

        static_energy_directory = directory / "static_energy"
        # run a static energy and bader/badelf analysis using the same structure
        # as above.
        # !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        # We need to make a new directory because only one vasp workflow can
        # be run in each directory.
        os.mkdir(static_energy_directory)
        shutil.copyfile(
            relaxation2_directory / "WAVECAR", static_energy_directory / "WAVECAR"
        )
        static_energy_result = cls.static_energy_workflow.run(
            structure=relaxation2_result,
            command=command,
            directory=static_energy_directory,
            # copy_previous_directory=True,
        )



# -*- coding: utf-8 -*-

from .badelf_hse import BadElf__Badelf__BadelfHse
from .badelf_pbesol import BadElf__Badelf__BadelfPbesol
from .prebadelf_dft import (
    StaticEnergy__Vasp__WarrenLabPrebadelfHse,
    StaticEnergy__Vasp__WarrenLabPrebadelfPbesol,
)
from .test import BadElf__Badelf__BadelfTest



# -*- coding: utf-8 -*-

from simmate.engine import S3Workflow


class PopulationAnalysis__Bader__Bader(S3Workflow):
    required_files = ["CHGCAR", "CHGCAR_sum", "POTCAR"]
    use_database = False
    use_previous_directory = ["CHGCAR", "CHGCAR_sum", "POTCAR"]
    parent_workflows = ["population-analysis.vasp-bader.bader-matproj"]

    command = "bader CHGCAR -ref CHGCAR_sum -b weight > bader.out"
    """
    The command to call the executable, which is typically bader. Note we
    use the `-b weight` by default, which means we apply the weight method for
    partitioning from of 
    [Yu and Trinkle](http://theory.cm.utexas.edu/henkelman/code/bader/download/yu11_064111.pdf).
    
    This command is modified to use the `-ref` file as the reference for determining
    zero-flux surfaces when partitioning the CHGCAR. 
    
    There are cases where also use structures that contain "empty atoms" in them.
    This is to help with partitioning of electrides, which possess electron 
    density that is not associated with any atomic orbital. For these cases,
    you will see files like "CHGCAR_w_empty_atoms" used in the command.
    """



# -*- coding: utf-8 -*-

# The order that we import these different modules is important to prevent
# circular imports errors, so we prevent isort from changing this file.
# isort: skip_file

from .from_images import Diffusion__Vasp__NebFromImagesMit
from .from_images_mvl_ci import Diffusion__Vasp__NebFromImagesMvlCi
from .from_endpoints import Diffusion__Vasp__NebFromEndpointsMit
from .single_path import Diffusion__Vasp__NebSinglePathMit
from .all_paths import Diffusion__Vasp__NebAllPathsMit



# -*- coding: utf-8 -*-

import importlib
import logging
import os
import sys
from functools import wraps

import requests

import simmate


def get_conda_env() -> str:
    """
    Grab the name of the activate conda environment and returns it as a string.
    If there is no env, then an empty string is returned.
    """
    # Check the list of python paths and grab the first path that has "envs" in it.
    # Assume we don't have a conda env until proven otherwise
    env_name = ""
    for path in sys.path:
        if "envs" in path:
            # split the path into individual folder names (os.sep gives / or \\)
            folders = path.split(os.sep)
            # the conda env name will the name immediately after the /envs.
            # example path is '/home/jacksund/anaconda3/envs/simmate_dev/lib/python3.10'
            # where we want the name simmate_dev here.
            env_name = folders[folders.index("envs") + 1]
            # once we have found this, we can exit the loop
            break

    return env_name


def get_latest_version() -> str:
    """
    Looks at the jacks/simmate repo and grabs the latest release version.
    """
    # Access the data via a web request
    response = requests.get(
        "https://api.github.com/repos/jacksund/simmate/releases/latest"
    )

    # load the version from the json response
    # [1:] simply removes the first letter "v" from something like "v1.2.3"
    latest_version = response.json()["tag_name"][1:]
    return latest_version


def check_if_using_latest_version(current_version=simmate.__version__):
    """
    Checks if there's a newer version by looking at the latest release on Github
    and comparing it to the currently installed version
    """
    latest_version = get_latest_version()

    if current_version != latest_version:
        logging.warning(
            "There is a new version of Simmate available. "
            f"You are currently using v{current_version} while v{latest_version} "
            "is the latest."
        )


def get_chemical_subsystems(chemical_system: str):
    """
    Given a chemical system, this returns all chemical systems that are also
    contained within it.

    For example, "Y-C" would return ["Y", "C", "C-Y"]. Note that the returned
    list has elements of a given system in alphabetical order (i.e. it gives
    "C-Y" and not "Y-C")

    #### Parameters

    - `chemical_system`:
        A chemical system of elements. Elements must be separated by dashes (-)

    #### Returns

    - `subsystems`:
        A list of chemical systems that make up the input chemical system.
    """

    # TODO: this code may be better located elsewhere. Maybe even as a method for
    # the Composition class or alternatively as a ChemicalSystem class.

    # I convert the system to a composition where the number of atoms dont
    # apply here. (e.g. "Ca-N" --> "Ca1 N1")
    from simmate.toolkit import Composition

    composition = Composition(chemical_system.replace("-", ""))

    return composition.chemical_subsystems


def chunk_list(full_list: list, chunk_size: int) -> list:
    """
    Yields successive n-sized chunks from a list.
    """
    for i in range(0, len(full_list), chunk_size):
        yield full_list[i : i + chunk_size]


def str_to_datatype(
    parameter: str,
    value: str,
    type_mappings: dict = {},
    strip_quotes: bool = False,
):
    """
    When given a parameter name and it's value as a string, this helper
    function will use the key (parameter) to determine how to convert the
    val string to the proper python datatype (int, float, bool, list...).
    A full mapping of parameters should be provided, but if a parameter is
    given that isn't mapped, the value will be leave left as a string.

    This is often ment to read to/from non-standard parameter file formats.
    For example, VASP's INCAR does not follow any standard (yaml, toml, json,
    etc.) so this function helps read values into python types.

    Example type_mapping using random VASP settings:
    ``` python
    type_mapping = {
        "LDAU": bool,
        "EDIFF": float,
        "NSW": int,
        "LDAUL": list[int],
        "LDAUU": list[float],
        "DIPOL": list[list[float]],
    }
    ```
    """

    # If the value is not a string, then assume we are already in the
    # correct format. Note, an incorrect format will throw an error
    # somewhere below, which may be tricky for beginners to traceback.
    if not isinstance(value, str):
        return value

    # next, try grabbing the type from mapping dictionary. If the parameter is
    # not mapped, then we assume it is a str.
    target_type = type_mappings.get(parameter, str)

    # Now that we know the target type to convert to, we can go through
    # decide how to handle converting the value

    if target_type == str:
        # assert type(value) == str
        if strip_quotes and ("'" in value or '"' in value):
            return value.strip("'").strip('"')
        return value

    elif target_type == int:
        # sometimes "1." was written to indicate an integer so check for
        # this and remove it if needed.
        if value[-1] == ".":
            value = value[:-1]
        # return the value integer
        return int(value)

    elif target_type == float:
        # return the value float
        return float(value)

    elif target_type == bool:
        # Python is weird where bool("FALSE") will return True... So I need
        # to convert the string to lowercase and read it to know what to
        # return here.
        if "t" in value.lower():
            return True
        elif "f" in value.lower():
            return False

    elif target_type == list[float]:
        final_list = []
        for item in value.split():
            # Sometimes, the values are given as "3*0.1 2*0.5" where the "*"
            # means to include that value that many times. For example, this
            # input would be the same as "0.1 0.1 0.1 0.5 0.5". We need to
            # account for this when parsing.
            if "*" in item:
                nsubitems, subitem = item.split("*")
                for n in range(int(nsubitems)):
                    final_list.append(float(subitem))
            else:
                final_list.append(float(item))
        return final_list

    elif target_type == list[int]:
        return [int(item) for item in value.split()]

    # BUG: I assume we want to split by spaces, but I don't know of a good
    # way to override this on a per-variable basis.
    elif target_type == list[str]:
        return value.split()

    # These vectors are always 3x floats
    elif target_type == list[list[float]]:
        # convert a string of...
        #   "x1 y1 z1 x2 y2 z2 x3 y3 z3"
        # to...
        #   [x1,y1,z1,x2,y2,z2,x3,y3,z3] (list of floats)
        # and then to...
        #   [[x1,y1,z1],[x2,y2,z2],[x3,y3,z3]]
        value = [float(item) for item in value.split()]
        return [value[i : i + 3] for i in range(0, len(value), 3)]

    # If it is not in the common keys listed, just leave it as a string.
    else:
        logging.warning(
            "Unknown parameter mapping of {parameter}: {target_type}. "
            "Leaving as str."
        )
        return value


def get_class(class_path: str):
    """
    Given the import path for a python class (e.g. path.to.MyClass), this
    utility will load the class given (MyClass).
    """
    config_modulename = ".".join(class_path.split(".")[:-1])
    config_name = class_path.split(".")[-1]
    config_module = importlib.import_module(config_modulename)
    config = getattr(config_module, config_name)
    return config


def get_app_submodule(
    app_name: str,
    submodule_name: str,
) -> str:
    """
    Checks if an app has a submodule present and returns the import path for it.
    This is useful for checking if there are workflows or urls defined, which
    are optional accross all apps. None is return if no app exists
    """
    config = get_class(app_name)
    app_path = config.name
    submodule_path = f"{app_path}.{submodule_name}"

    # check if there is a workflows module in the app, and if so,
    # try loading the workflows.
    #   stackoverflow.com/questions/14050281
    has_submodule = importlib.util.find_spec(submodule_path) is not None

    return submodule_path if has_submodule else None


def bypass_nones(bypass_kwarg: str = None, multi_cols: bool = False):
    """
    experimental utility that removes None values before passing a list of
    entries to a method or function. The method or function then returns
    a list of results with the None values placed back in the proper index.
    """
    # https://stackoverflow.com/questions/5929107/decorators-with-parameters

    def decorator(function_to_wrap):
        @wraps(function_to_wrap)
        def wrapper(*args, **kwargs):
            # breakpoint()

            # grab the list of original inputs
            entries = kwargs.pop(bypass_kwarg)
            # BUG: position arguments will cause errors

            # remove None values and keep a record of their original positions
            passed_entries = []
            failed_idxs = []
            for idx, entry in enumerate(entries):
                if entry:
                    passed_entries.append(entry)
                else:
                    failed_idxs.append(idx)
            kwargs[bypass_kwarg] = passed_entries

            # breakpoint()

            # RUN THE ORIGINAL METHOD
            results_orig = function_to_wrap(*args, **kwargs)

            # breakpoint()

            if not multi_cols:
                results_orig = [results_orig]
            results_final = []
            for column in results_orig:
                # Add back None values in the proper position for this single column
                col_final = []
                failed_count = 0
                for idx, result in enumerate(column):
                    while (idx + failed_count) in failed_idxs:
                        failed_count += 1
                        col_final.append(None)
                    col_final.append(result)
                # there may be extra None values needed at the end. We add Nones until
                # we get the correct list length
                while len(entries) != len(col_final):
                    col_final.append(None)
                results_final.append(col_final)

            return results_final if multi_cols else results_final[0]

        return wrapper

    return decorator


class dotdict(dict):
    """
    Provides dot.notation access to dictionary attributes
    """

    # This class is modified from suggestions here:
    # https://stackoverflow.com/questions/2352181/

    # BUG: these two methods need to be updated to be recursive like I did
    # with __getattr__ below
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

    def __getattr__(self, name: str):
        if name not in self.keys():
            raise Exception(f"Unknown property: {name}")

        setting = self.get(name, None)

        # if the property accessed is a dictionary, then we make it a dotdict
        # so that we can perform recursive dot access
        if isinstance(setting, dict):
            return dotdict(setting)
        else:
            return setting


def deep_update(default_dict: dict, override_dict: dict) -> dict:
    """
    Recursively update the default dictionary with values from the override dictionary.
    """

    # This fxn mirrors pydantic's deep_update util
    # https://stackoverflow.com/questions/3232943/

    final_dict = default_dict.copy()

    for key, value in override_dict.items():
        if (
            isinstance(value, dict)
            and key in final_dict
            and isinstance(final_dict[key], dict)
        ):
            final_dict[key] = deep_update(final_dict[key], value)

        else:
            final_dict[key] = value

    return final_dict


def get_docker_command(
    image: str,
    entrypoint: str = None,
    volumes: list[str] = [],
) -> str:
    """
    Given common input parameters, this util builds a `docker run` command
    and returns it as a string.We then let other functionality
    (such as S3Workflows) handle running the command.

    For advanced docker cases, use the `docker-py` package instead.
    """
    command = "docker run "

    if entrypoint:
        command += f"--entrypoint {entrypoint} "

    for volume in volumes:
        command += f"--volume {volume} "

    command += image

    return command



# -*- coding: utf-8 -*-

"""

VoxelData Converters
--------------------

Simmate still focuses almost exclusively on bulk crystal structures, so this module is empty at the moment. However, we keep this here for outlining and planning purposes. In the future, this will be a library of converters for volume-based data files and object types.

"""



# -*- coding: utf-8 -*-

from simmate.apps.quantum_espresso.workflows.base import PwscfWorkflow


class StaticEnergy__QuantumEspresso__Quality00(PwscfWorkflow):
    """
    A very a basic static energy calc for testing
    """

    control = dict(
        pseudo_dir__auto=True,
        restart_mode="from_scratch",
        calculation="scf",
        tstress=True,
        tprnfor=True,
    )

    system = dict(
        ibrav=0,
        nat__auto=True,
        ntyp__auto=True,
        ecutwfc__auto="efficiency",
        ecutrho__auto="efficiency",
    )

    electrons = dict(
        diagonalization="cg",
        mixing_mode="plain",
        mixing_beta=0.7,
        conv_thr="1.0e-8",
    )

    psuedo_mappings_set = "SSSP_PBE_EFFICIENCY"

    k_points = dict(
        spacing=0.5,
        gamma_centered=True,
    )



# -*- coding: utf-8 -*-

import typer

setup_app = typer.Typer(rich_markup_mode="markdown")


@setup_app.callback(no_args_is_help=True)
def setup():
    """
    A group of commands for configuring QE and PWscf
    """
    pass


@setup_app.command()
def sssp():
    """
    Downloads and configures potentials from SSSP
    """
    from simmate.apps.quantum_espresso.inputs.potentials_sssp import setup_sssp

    setup_sssp()



# -*- coding: utf-8 -*-

import logging
from pathlib import Path

from rich.progress import track

from simmate.toolkit import Structure
from simmate.utilities import get_directory


def write_and_submit_structures(
    structures: list[Structure],
    foldername: Path,
    workflow,  # cant set type because a database module depends on this method
    workflow_kwargs: dict,
):
    if not structures:
        return []

    logging.info("Writing CIFs and submitting structures")

    # sort structure by increasing size (smallest first)
    structures.sort(key=lambda s: s.num_sites)

    # disable the logs while we submit
    logger = logging.getLogger()
    logger.disabled = True

    nalready_submitted = 0
    directory = get_directory(foldername)
    states = []
    for i, s in enumerate(track(structures)):
        # check if the structure has been submitted before, and if so, skip it
        if workflow.all_results.filter(source=s.source).exists():
            nalready_submitted += 1
            continue

        i_cleaned = str(i).zfill(3)  # converts 1 to 001
        s.to(filename=str(directory / f"{i_cleaned}.cif"), fmt="cif")

        state = workflow.run_cloud(
            structure=s,
            **workflow_kwargs,
        )
        states.append(state)

    logger.disabled = False

    if nalready_submitted > 0:
        logging.info(
            f"{nalready_submitted} structures were already submitted "
            "and therefore skipped."
        )

    return states
